import unittest
import tempfile

from pycoin.encoding.hexbytes import h2b

from .ToolTest import ToolTest


class BlockTest(ToolTest):

    def test_block_dump(self):
        block_hex = (
            '01000000bddd99ccfda39da1b108ce1a5d70038d0a967bacb68b6b63065f626a000000'
            '0044f672226090d85db9a9f2fbfe5f0f9609b387af7be5b7fbb7a1767c831c9e995dbe'
            '6649ffff001d05e0ed6d01010000000100000000000000000000000000000000000000'
            '00000000000000000000000000ffffffff0704ffff001d010effffffff0100f2052a01'
            '00000043410494b9d3e76c5b1629ecf97fff95d7a4bbdac87cc26099ada28066c6ff1e'
            'b9191223cd897194a08d0c2726c5747f1db49e8cf90e75dc3e3550ae9b30086f3cd5aa'
            'ac00000000')
        block_bin = h2b(block_hex)
        block_file = tempfile.NamedTemporaryFile()
        block_file.write(block_bin)
        block_file.flush()
        output = self.launch_tool("block %s" % block_file.name)
        self.assertEqual(output, """215 bytes   block hash 0000000082b5015589a3fdf2d4baff403e6f0be035a5d9742c1cae6295464449
version 1
prior block hash 000000006a625f06636b8bb6ac7b960a8d03705d1ace08b1a19da3fdcc99ddbd
merkle root 44f672226090d85db9a9f2fbfe5f0f9609b387af7be5b7fbb7a1767c831c9e99
timestamp 2009-01-09T03:02:53
difficulty 486604799
nonce 1844305925
1 transaction
Tx #0:
Version:  1  tx hash 999e1c837c76a1b7fbb7e57baf87b309960f5ffefbf2a9b95dd890602272f644  134 bytes
TxIn count: 1; TxOut count: 1
Lock time: 0 (valid anytime)
Input:
   0: COINBASE   50000.00000 mBTC
Output:
   0: 1FvzCLoTPGANNjWoUo6jUGuAG3wg1w4YjR receives  50000.00000 mBTC
Total input   50000.00000 mBTC
Total output  50000.00000 mBTC
Total fees        0.00000 mBTC

""")


def main():
    unittest.main()


if __name__ == "__main__":
    main()



from ..encoding.sec import is_sec, public_pair_to_hash160_sec, sec_to_public_pair, EncodingError

from pycoin.coins.SolutionChecker import ScriptError
from pycoin.satoshi.checksigops import parse_signature_blob
from pycoin.satoshi.der import UnexpectedDER


class WhoSigned(object):
    def __init__(self, script_tools, address_api, generator):
        self._script_tools = script_tools
        self._address = address_api
        self._generator = generator

    def solution_blobs(self, tx, tx_in_idx):
        """
        This iterator yields data blobs that appear in the the last solution_script or the witness.
        """
        sc = tx.SolutionChecker(tx)
        tx_context = sc.tx_context_for_idx(tx_in_idx)
        # set solution_stack in case there are no results from puzzle_and_solution_iterator
        solution_stack = []
        for puzzle_script, solution_stack, flags, sighash_f in sc.puzzle_and_solution_iterator(tx_context):
            pass
            # we only care about the last one
        for s in solution_stack:
            yield s

    def extract_signatures(self, tx, tx_in_idx):
        sc = tx.SolutionChecker(tx)
        tx_context = sc.tx_context_for_idx(tx_in_idx)
        # set solution_stack in case there are no results from puzzle_and_solution_iterator
        solution_stack = []
        for puzzle_script, solution_stack, flags, sighash_f in sc.puzzle_and_solution_iterator(tx_context):
            pass
            # we only care about the last one

        vm = sc.VM(puzzle_script, tx_context, sighash_f, flags=flags, initial_stack=solution_stack[:])
        for data in self.solution_blobs(tx, tx_in_idx):
            try:
                sig_pair, sig_type = parse_signature_blob(data)
                sig_hash = sighash_f(sig_type, sig_blobs=[], vm=vm)
                yield (data, sig_hash)
            except (ValueError, TypeError, UnexpectedDER, ScriptError):
                continue

    def extract_secs(self, tx, tx_in_idx):
        """
        For a given script solution, iterate yield its sec blobs
        """
        sc = tx.SolutionChecker(tx)
        tx_context = sc.tx_context_for_idx(tx_in_idx)
        # set solution_stack in case there are no results from puzzle_and_solution_iterator
        solution_stack = []
        for puzzle_script, solution_stack, flags, sighash_f in sc.puzzle_and_solution_iterator(tx_context):
            for opcode, data, pc, new_pc in self._script_tools.get_opcodes(puzzle_script):
                if data and is_sec(data):
                    yield data
            for data in solution_stack:
                if is_sec(data):
                    yield data

    def public_pairs_for_script(self, tx, tx_in_idx, generator):
        """
        For a given script, iterate over and pull out public pairs encoded as sec values.
        """
        public_pairs = []
        for sec in self.extract_secs(tx, tx_in_idx):
            try:
                public_pairs.append(sec_to_public_pair(sec, generator))
            except EncodingError:
                pass
        return public_pairs

    def public_pairs_signed(self, tx, tx_in_idx):
        signed_by = []

        public_pairs = self.public_pairs_for_script(tx, tx_in_idx, self._generator)

        for signature, sig_hash in self.extract_signatures(tx, tx_in_idx):
            sig_pair, sig_type = parse_signature_blob(signature)

            for public_pair in public_pairs:
                if self._generator.verify(public_pair, sig_hash, sig_pair):
                    signed_by.append((public_pair, sig_pair, sig_type))
        return signed_by

    def who_signed_tx(self, tx, tx_in_idx):
        """
        Given a transaction (tx) an input index (tx_in_idx), attempt to figure
        out which addresses where used in signing (so far). This method
        depends on tx.unspents being properly configured. This should work on
        partially-signed MULTISIG transactions (it will return as many
        addresses as there are good signatures).
        Returns a list of (public_pairs, sig_type) pairs.
        """
        public_pair_sig_type_list = self.public_pairs_signed(tx, tx_in_idx)
        sig_type_list = [pp[-1] for pp in public_pair_sig_type_list]
        hash160_list = [public_pair_to_hash160_sec(pp[0]) for pp in public_pair_sig_type_list]
        address_list = [self._address.for_p2pkh(h160) for h160 in hash160_list]
        return list(zip(address_list, sig_type_list))



import itertools
import unittest

from pycoin.symbols.btc import network


Tx = network.tx


class MultisigIndividualTest(unittest.TestCase):
    def multisig_M_of_N_individually(self, M, N):
        keys = [network.keys.private(secret_exponent=i) for i in range(1, N+2)]
        tx_in = Tx.TxIn.coinbase_tx_in(script=b'')
        script = network.contract.for_multisig(m=M, sec_keys=[key.sec() for key in keys[:N]])
        tx_out = Tx.TxOut(1000000, script)
        tx1 = Tx(version=1, txs_in=[tx_in], txs_out=[tx_out])
        for partial_key_list in itertools.permutations(keys[:N], M):
            tx2 = network.tx_utils.create_tx(tx1.tx_outs_as_spendable(), [keys[-1].address()])
            for key in partial_key_list:
                self.assertEqual(tx2.bad_solution_count(), 1)
                hash160_lookup = network.tx.solve.build_hash160_lookup([key.secret_exponent()])
                tx2.sign(hash160_lookup=hash160_lookup)
            self.assertEqual(tx2.bad_solution_count(), 0)

    def test_multisig_one_at_a_time(self):
        for N in range(1, 4):
            for M in range(1, N+1):
                self.multisig_M_of_N_individually(M, N)


if __name__ == "__main__":
    unittest.main()



import unittest


from pycoin.crack.bip32 import ascend_bip32, crack_bip32
from pycoin.symbols.btc import network


class CrackBIP32Test(unittest.TestCase):

    def setUp(self):
        self.bip32_key = network.keys.bip32_seed(b"foo")

    def test_crack_bip32(self):
        bip32_key = self.bip32_key
        bip32_pub = bip32_key.public_copy()
        secret_exponent_p0_1_7_9 = bip32_key.subkey_for_path("0/1/7/9").secret_exponent()
        cracked_bip32_node = crack_bip32(bip32_pub, secret_exponent_p0_1_7_9, "0/1/7/9")
        self.assertEqual(cracked_bip32_node.hwif(as_private=True), bip32_key.hwif(as_private=True))

    def test_ascend_bip32(self):
        bip32_key = self.bip32_key
        bip32_pub = bip32_key.public_copy()
        secret_exponent_p9 = bip32_key.subkey_for_path("9").secret_exponent()
        secret_exponent = ascend_bip32(bip32_pub, secret_exponent_p9, 9)
        self.assertEqual(secret_exponent, bip32_key.secret_exponent())



#!/usr/bin/env python

# This script creates a fake coinbase transaction to an address of your
# choosing so you can test code that spends this output.

import sys

from pycoin.symbols.btc import network


def main():
    if len(sys.argv) != 2:
        print("usage: %s address" % sys.argv[0])
        sys.exit(-1)

    # validate the address
    address = sys.argv[1]
    assert network.parse.address(address) is not None

    print("creating coinbase transaction to %s" % address)

    tx_in = network.tx.TxIn.coinbase_tx_in(script=b'')
    tx_out = network.tx.TxOut(50*1e8, network.contract.for_address(address))
    tx = network.tx(1, [tx_in], [tx_out])
    print("Here is the tx as hex:\n%s" % tx.as_hex())


if __name__ == '__main__':
    main()



from pycoin.satoshi.flags import SIGHASH_FORKID

from ..bitcoin.SolutionChecker import BitcoinSolutionChecker


class BcashSolutionChecker(BitcoinSolutionChecker):
    def _signature_hash(self, tx_out_script, unsigned_txs_out_idx, hash_type):
        """
        Return the canonical hash for a transaction. We need to
        remove references to the signature, since it's a signature
        of the hash before the signature is applied.

        tx_out_script: the script the coins for unsigned_txs_out_idx are coming from
        unsigned_txs_out_idx: where to put the tx_out_script
        hash_type: one of SIGHASH_NONE, SIGHASH_SINGLE, SIGHASH_ALL,
        optionally bitwise or'ed with SIGHASH_ANYONECANPAY
        """

        if hash_type & SIGHASH_FORKID != SIGHASH_FORKID:
            raise self.ScriptError()

        return self._signature_for_hash_type_segwit(tx_out_script, unsigned_txs_out_idx, hash_type)



from pycoin.networks.bitcoinish import create_bitcoinish_network

network = create_bitcoinish_network(
    symbol="FTX", network_name="Feathercoin", subnet_name="testnet",
    wif_prefix_hex="c1", address_prefix_hex="41", pay_to_script_prefix_hex="c4",
    bip32_prv_prefix_hex="04358394", bip32_pub_prefix_hex="043587cf")



from . import errno
from pycoin.coins.SolutionChecker import ScriptError

from .flags import VERIFY_MINIMALDATA


def do_OP_VERIFY(vm):
    v = vm.bool_from_script_bytes(vm.pop())
    if not v:
        raise ScriptError("VERIFY failed", errno.VERIFY)


def do_OP_DEPTH(vm):
    """
    >>> s = [1, 2, 1, 2, 1, 2]
    >>> do_OP_DEPTH(s)
    >>> print(s)
    [1, 2, 1, 2, 1, 2, b'\\x06']
    """
    vm.push_int(len(vm.stack))


def do_OP_PICK(vm):
    """
    >>> s = [b'a', b'b', b'c', b'd', b'\2']
    >>> do_OP_PICK(s, require_minimal=True)
    >>> print(s)
    [b'a', b'b', b'c', b'd', b'b']
    """
    v = vm.pop_nonnegative()
    vm.append(vm[-v-1])


def do_OP_ROLL(vm):
    """
    >>> s = [b'a', b'b', b'c', b'd', b'\2']
    >>> do_OP_ROLL(s, require_minimal=True)
    >>> print(s)
    [b'a', b'c', b'd', b'b']
    """
    v = vm.pop_nonnegative()
    vm.append(vm.pop(-v-1))


def do_OP_SUBSTR(vm):
    """
    >>> s = [b'abcdef', b'\3', b'\2']
    >>> do_OP_SUBSTR(s, require_minimal=True)
    >>> print(s)
    [b'de']
    """
    pos = vm.pop_nonnegative()
    length = vm.pop_nonnegative()
    vm.append(vm.pop()[length:length+pos])


def do_OP_LEFT(vm):
    """
    >>> s = [b'abcdef', b'\3']
    >>> do_OP_LEFT(s, require_minimal=True)
    >>> print(len(s)==1 and s[0]==b'abc')
    True
    >>> s = [b'abcdef', b'']
    >>> do_OP_LEFT(s, require_minimal=True)
    >>> print(len(s) ==1 and s[0]==b'')
    True
    """
    pos = vm.pop_nonnegative()
    vm.append(vm.pop()[:pos])


def do_OP_RIGHT(vm):
    """
    >>> s = [b'abcdef', b'\\3']
    >>> do_OP_RIGHT(s, require_minimal=True)
    >>> print(s==[b'def'])
    True
    >>> s = [b'abcdef', b'\\0']
    >>> do_OP_RIGHT(s, require_minimal=False)
    >>> print(s==[b''])
    True
    """
    pos = vm.pop_nonnegative()
    if pos > 0:
        vm.append(vm.pop()[-pos:])
    else:
        vm.pop()
        vm.append(b'')


def do_OP_SIZE(vm):
    """
    >>> import binascii
    >>> s = [b'abcdef']
    >>> do_OP_SIZE(s)
    >>> print(s == [b'abcdef', b'\x06'])
    True
    >>> s = [b'abcdef'*1000]
    >>> do_OP_SIZE(s)
    >>> print(binascii.hexlify(s[-1]) == b'7017')
    True
    """
    vm.push_int(len(vm[-1]))


def do_OP_EQUAL(vm):
    """
    >>> s = [b'string1', b'string1']
    >>> do_OP_EQUAL(s)
    >>> print(s == [b'\1'])
    True
    >>> s = [b'string1', b'string2']
    >>> do_OP_EQUAL(s)
    >>> print(s == [b''])
    True
    """
    v1, v2 = [vm.pop() for i in range(2)]
    vm.append(vm.bool_to_script_bytes(v1 == v2))


def do_OP_EQUALVERIFY(vm):
    do_OP_EQUAL(vm)
    v = vm.bool_from_script_bytes(vm.pop())
    if not v:
        raise ScriptError("VERIFY failed", errno.EQUALVERIFY)


def pop_check_bounds(vm):
    if len(vm[-1]) > 4:
        raise ScriptError("overflow in binop", errno.UNKNOWN_ERROR)
    return vm.pop_int()


def make_bin_op(binop):
    def f(vm):
        v1, v2 = [pop_check_bounds(vm) for i in range(2)]
        vm.push_int(binop(v2, v1))
    return f


def make_bool_bin_op(binop):
    def f(vm):
        v1, v2 = [pop_check_bounds(vm) for i in range(2)]
        vm.append(vm.bool_to_script_bytes(binop(v2, v1)))
    return f


do_OP_ADD = make_bin_op(lambda x, y: x + y)
do_OP_SUB = make_bin_op(lambda x, y: x - y)
do_OP_MUL = make_bin_op(lambda x, y: x * y)
do_OP_DIV = make_bin_op(lambda x, y: x // y)
do_OP_MOD = make_bin_op(lambda x, y: x % y)
do_OP_LSHIFT = make_bin_op(lambda x, y: x << y)
do_OP_RSHIFT = make_bin_op(lambda x, y: x >> y)
do_OP_BOOLAND = make_bool_bin_op(lambda x, y: x and y)
do_OP_BOOLOR = make_bool_bin_op(lambda x, y: x or y)
do_OP_NUMEQUAL = make_bool_bin_op(lambda x, y: x == y)
do_OP_NUMNOTEQUAL = make_bool_bin_op(lambda x, y: x != y)
do_OP_LESSTHAN = make_bool_bin_op(lambda x, y: x < y)
do_OP_GREATERTHAN = make_bool_bin_op(lambda x, y: x > y)
do_OP_LESSTHANOREQUAL = make_bool_bin_op(lambda x, y: x <= y)
do_OP_GREATERTHANOREQUAL = make_bool_bin_op(lambda x, y: x >= y)
do_OP_MIN = make_bin_op(min)
do_OP_MAX = make_bin_op(max)


def do_OP_NUMEQUALVERIFY(vm):
    do_OP_NUMEQUAL(vm)
    do_OP_VERIFY(vm)


def do_OP_WITHIN(vm):
    """
    >>> s = [b'b', b'a', b'c']
    >>> do_OP_WITHIN(s, False)
    >>> print(s == [b'\1'])
    True
    >>> s = [b'd', b'a', b'c']
    >>> do_OP_WITHIN(s, False)
    >>> print(s == [b''])
    True
    """
    v3, v2, v1 = [vm.pop_int() for i in range(3)]
    ok = (v2 <= v1 < v3)
    vm.append(vm.bool_to_script_bytes(ok))


def make_unary_num_op(unary_f):
    def f(vm):
        vm.push_int(unary_f(pop_check_bounds(vm)))
    return f


do_OP_1ADD = make_unary_num_op(lambda x: x + 1)
do_OP_1SUB = make_unary_num_op(lambda x: x - 1)
do_OP_2MUL = make_unary_num_op(lambda x: x << 1)
do_OP_2DIV = make_unary_num_op(lambda x: x >> 1)
do_OP_NEGATE = make_unary_num_op(lambda x: -x)
do_OP_ABS = make_unary_num_op(lambda x: abs(x))


def do_OP_NOT(vm):
    return vm.append(vm.bool_to_script_bytes(not pop_check_bounds(vm)))


def do_OP_0NOTEQUAL(vm):
    return vm.push_int(
        vm.bool_from_script_bytes(
            vm.pop(), require_minimal=vm.flags & VERIFY_MINIMALDATA))


"""
The MIT License (MIT)

Copyright (c) 2013 by Richard Kiss

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""



import functools

from pycoin.encoding.hexbytes import b2h_rev
from pycoin.satoshi.satoshi_struct import parse_struct, stream_struct

ITEM_TYPE_TX = 1
ITEM_TYPE_BLOCK = 2
ITEM_TYPE_MERKLEBLOCK = 3
INV_CMPCT_BLOCK = 4

INV_WITNESS_FLAG = 1 << 30
INV_TYPE_MASK = 0xffffffff >> 2


@functools.total_ordering
class InvItem(object):
    def __init__(self, item_type, data, dont_check=False):
        if not dont_check:
            assert item_type in (ITEM_TYPE_TX, ITEM_TYPE_BLOCK, ITEM_TYPE_MERKLEBLOCK)
        self.item_type = item_type
        assert isinstance(data, bytes)
        assert len(data) == 32
        self.data = data

    def __str__(self):
        INV_TYPES = ["?", "Tx", "Block", "Merkle"]
        idx = self.item_type
        if not 0 < idx < 4:
            idx = 0
        return "InvItem %s [%s]" % (INV_TYPES[idx], b2h_rev(self.data))

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash((self.item_type, self.data))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.item_type == other.item_type and self.data == other.data
        return False

    def __lt__(self, other):
        return (self.item_type, self.data) < (other.item_type, other.data)

    def stream(self, f):
        stream_struct("L#", f, self.item_type, self.data)

    @classmethod
    def parse(self, f):
        return self(*parse_struct("L#", f), dont_check=True)



import re
import threading
import warnings

from pycoin.networks.default import get_current_netcode

from .bitcoind import BitcoindProvider
from .blockexplorer import BlockExplorerProvider
from .blockchain_info import BlockchainInfoProvider
from .blockcypher import BlockcypherProvider
from .chain_so import ChainSoProvider
from .insight import InsightProvider
from .btgexp import BTGExpProvider

from .env import main_cache_dir, config_string_for_netcode_from_env
from .env import tx_read_cache_dirs, tx_writable_cache_dir
from .tx_db import TxDb


THREAD_LOCALS = threading.local()


# PYCOIN_BTC_PROVIDERS="blockchain.info blockexplorer.com blockcypher.com chain.so"
# PYCOIN_BTC_PROVIDERS="insight:http(s?)://hostname/url bitcoinrpc://user:passwd@hostname:8332"


def service_provider_methods(method_name, service_providers):
    methods = [getattr(m, method_name, None) for m in service_providers]
    methods = [m for m in methods if m]
    return methods


def spendables_for_address(address, netcode, format=None):
    """
    Return a list of Spendable objects for the
    given bitcoin address.

    Set format to "text" or "dict" to transform return value
    from an object to a string or dict.

    This is intended to be a convenience function. There is no way to know that
    the list returned is a complete list of spendables for the address in question.

    You can verify that they really do come from the existing transaction
    by calling tx_utils.validate_unspents.
    """
    if format:
        method = "as_%s" % format
    for m in service_provider_methods("spendables_for_address", get_default_providers_for_netcode(netcode)):
        try:
            spendables = m(address)
            if format:
                spendables = [getattr(s, method)() for s in spendables]
            return spendables
        except Exception:
            pass
    return []


def get_tx_db(netcode=None):
    lookup_methods = service_provider_methods("tx_for_tx_hash", get_default_providers_for_netcode(netcode))
    read_cache_dirs = tx_read_cache_dirs()
    writable_cache_dir = tx_writable_cache_dir()
    return TxDb(lookup_methods=lookup_methods, read_only_paths=read_cache_dirs,
                writable_cache_path=writable_cache_dir)


def message_about_tx_cache_env():
    if main_cache_dir() is None:
        return "consider setting environment variable PYCOIN_CACHE_DIR=~/.pycoin_cache to"\
               " cache transactions fetched via web services"


def all_providers_message(method, netcode):
    if len(service_provider_methods(method, get_default_providers_for_netcode(netcode))) == 0:
        return "no service providers found for %s; consider setting environment variable "\
            "PYCOIN_%s_PROVIDERS" % (method, netcode)


def message_about_spendables_for_address_env(netcode):
    return all_providers_message("spendables_for_address", netcode)


def message_about_tx_for_tx_hash_env(netcode):
    return all_providers_message("tx_for_tx_hash", netcode)


def bitcoin_rpc_init(match, netcode):
    username, password, hostname, port = match.group("user", "password", "hostname", "port")
    return BitcoindProvider("http://%s:%s@%s:%s" % (username, password, hostname, port))


def insight_init(match, netcode):
    return InsightProvider(base_url=match.group("url"), netcode=netcode)


DESCRIPTOR_CRE_INIT_TUPLES = [
    (re.compile(
        r"^bitcoinrpc://(?P<user>\S*):(?P<password>\S*)\@(?P<hostname>\S*)(:(?P<port>\d*))"),
        bitcoin_rpc_init),
    (re.compile(r"^blockchain\.info$"), lambda m, netcode: BlockchainInfoProvider(netcode)),
    (re.compile(r"^blockcypher\.com$"), lambda m, netcode: BlockcypherProvider(netcode)),
    (re.compile(r"^blockexplorer\.com$"), lambda m, netcode: BlockExplorerProvider(netcode)),
    (re.compile(r"^chain\.so$"), lambda m, netcode: ChainSoProvider(netcode)),
    (re.compile(r"^insight:(?P<url>\S*)$"), insight_init),
    (re.compile(r"^btgexp.com"), lambda m, netcode: BTGExpProvider()),
]


def provider_for_descriptor_and_netcode(descriptor, netcode=None):
    if netcode is None:
        netcode = get_current_netcode()
    for cre, f in DESCRIPTOR_CRE_INIT_TUPLES:
        m = cre.match(descriptor)
        if m:
            return f(m, netcode)
    return None


def providers_for_config_string(config_string, netcode):
    providers = []
    for d in config_string.split():
        p = provider_for_descriptor_and_netcode(d, netcode)
        if p:
            providers.append(p)
        else:
            warnings.warn("can't parse provider %s in config string" % d)
    return providers


def providers_for_netcode_from_env(netcode):
    return providers_for_config_string(config_string_for_netcode_from_env(netcode), netcode)


def get_default_providers_for_netcode(netcode=None):
    if netcode is None:
        netcode = get_current_netcode()
    if not hasattr(THREAD_LOCALS, "providers"):
        THREAD_LOCALS.providers = {}
    if netcode not in THREAD_LOCALS.providers:
        THREAD_LOCALS.providers[netcode] = providers_for_netcode_from_env(netcode)
    return THREAD_LOCALS.providers[netcode]


def set_default_providers_for_netcode(netcode, provider_list):
    if not hasattr(THREAD_LOCALS, "providers"):
        THREAD_LOCALS.providers = {}
    THREAD_LOCALS.providers[netcode] = provider_list



"""
Provide the following functions, all cribbed from the library
`six <http://pythonhosted.org/six/>`_.

iterbytes(buf):
    return an iterator of ints corresponding to the bytes of buf

indexbytes(buf, i):
    return the int for the ith byte of buf

int2byte(an_int):
    convert a small integer (< 256) into bytes (with length 1)

byte2int(bs):
    turn bs[0] into an int (0-255)
"""

import functools
import itertools
import operator
import struct

if bytes == str:
    iterbytes = functools.partial(itertools.imap, ord)

    def indexbytes(buf, i):
        return ord(buf[i])
    int2byte = chr

    def byte2int(bs):
        return ord(bs[0])
else:
    iterbytes = iter
    indexbytes = operator.getitem
    int2byte = struct.Struct(">B").pack
    byte2int = operator.itemgetter(0)



import json
from .agent import urlopen

from pycoin.coins.bitcoin.Tx import Tx
from pycoin.encoding.hexbytes import b2h_rev


class BlockExplorerProvider(object):
    def __init__(self, netcode):
        url_stub = {"BTC": "blockexplorer.com", "XTN": "testnet.blockexplorer.com"}.get(netcode)
        if url_stub is None:
            raise ValueError("unsupported netcode %s" % netcode)
        self.url = "https://%s/api" % url_stub

    def tx_for_tx_hash(self, tx_hash):
        """
        Get a Tx by its hash.
        """
        url = "%s/rawtx/%s" % (self.url, b2h_rev(tx_hash))
        d = urlopen(url).read()
        j = json.loads(d.decode("utf8"))
        tx = Tx.from_hex(j.get("rawtx", ""))
        if tx.hash() == tx_hash:
            return tx



from ..encoding.sec import sec_to_public_pair, EncodingError
from ..intbytes import byte2int, indexbytes, iterbytes

from . import der
from . import errno

from .flags import (
    SIGHASH_ALL, SIGHASH_SINGLE, SIGHASH_ANYONECANPAY,
    VERIFY_NULLDUMMY, VERIFY_NULLFAIL, VERIFY_STRICTENC,
    VERIFY_DERSIG, VERIFY_LOW_S, VERIFY_WITNESS_PUBKEYTYPE
)

from pycoin.coins.SolutionChecker import ScriptError


def _check_valid_signature_1(sig):
    ls = len(sig)
    if ls < 9 or ls > 73:
        raise ScriptError("bad signature size", errno.SIG_DER)
    if sig[0] != 0x30:
        raise ScriptError("bad signature byte 0", errno.SIG_DER)
    if sig[1] != ls - 3:
        raise ScriptError("signature size wrong", errno.SIG_DER)
    r_len = sig[3]
    if 5 + r_len >= ls:
        raise ScriptError("r length exceed signature size", errno.SIG_DER)


def _check_valid_signature_2(sig):
    ls = len(sig)
    r_len = sig[3]
    s_len = sig[5 + r_len]
    if r_len + s_len + 7 != ls:
        raise ScriptError("r and s size exceed signature size", errno.SIG_DER)
    if sig[2] != 2:
        raise ScriptError("R value region does not start with 0x02", errno.SIG_DER)
    if r_len == 0:
        raise ScriptError("zero-length R value", errno.SIG_DER)
    if sig[4] & 0x80:
        raise ScriptError("sig R value not allowed to be negative", errno.SIG_DER)
    if r_len > 1 and sig[4] == 0 and not (sig[5] & 0x80):
        raise ScriptError(
            "R value can't have leading 0 byte unless doing so would make it negative", errno.SIG_DER)
    if sig[r_len + 4] != 2:
        raise ScriptError("S value region does not start with 0x02", errno.SIG_DER)
    if s_len == 0:
        raise ScriptError("zero-length S value", errno.SIG_DER)
    if sig[r_len + 6] & 0x80:
        raise ScriptError("negative S values not allowed", errno.SIG_DER)
    if s_len > 1 and sig[r_len + 6] == 0 and not (sig[r_len + 7] & 0x80):
        raise ScriptError(
            "S value can't have leading 0 byte unless doing so would make it negative", errno.SIG_DER)


def check_valid_signature(sig):
    # ported from bitcoind src/script/interpreter.cpp IsValidSignatureEncoding
    sig = [s for s in iterbytes(sig)]
    _check_valid_signature_1(sig)
    _check_valid_signature_2(sig)


def check_low_der_signature(sig_pair, generator):
    # IsLowDERSignature
    r, s = sig_pair
    hi_s = generator.p() - s
    if hi_s < s:
        raise ScriptError("signature has high S value", errno.SIG_HIGH_S)


def check_defined_hashtype_signature(sig):
    # IsDefinedHashtypeSignature
    if len(sig) == 0:
        raise ScriptError("signature is length 0")
    hash_type = indexbytes(sig, -1) & (~SIGHASH_ANYONECANPAY)
    if hash_type < SIGHASH_ALL or hash_type > SIGHASH_SINGLE:
        raise ScriptError("bad hash type after signature", errno.SIG_HASHTYPE)


def parse_signature_blob(sig_blob):
    if len(sig_blob) == 0:
        raise ValueError("empty sig_blob")
    sig_pair = der.sigdecode_der(sig_blob[:-1], use_broken_open_ssl_mechanism=True)
    signature_type = ord(sig_blob[-1:])
    return sig_pair, signature_type


def parse_and_check_signature_blob(sig_blob, flags, vm):
    if len(sig_blob) == 0:
        raise ValueError("empty sig_blob")
    if flags & (VERIFY_DERSIG | VERIFY_LOW_S | VERIFY_STRICTENC):
        check_valid_signature(sig_blob)
    if flags & VERIFY_STRICTENC:
        check_defined_hashtype_signature(sig_blob)
    sig_pair, signature_type = parse_signature_blob(sig_blob)
    if flags & VERIFY_LOW_S:
        generator = vm.generator_for_signature_type(signature_type)
        check_low_der_signature(sig_pair, generator)
    return sig_pair, signature_type


def check_public_key_encoding(blob):
    lb = len(blob)
    if lb >= 33:
        fb = byte2int(blob)
        if fb == 4:
            if lb == 65:
                return
        elif fb in (2, 3):
            if lb == 33:
                return
    raise ScriptError("invalid public key blob", errno.PUBKEYTYPE)


def checksig(vm, sig_pair, signature_type, pair_blob, blobs_to_delete,
             sighash_cache, verify_witness_pubkeytype, verify_strict):
    generator = vm.generator_for_signature_type(signature_type)
    if verify_strict:
        check_public_key_encoding(pair_blob)
    if verify_witness_pubkeytype:
        if byte2int(pair_blob) not in (2, 3) or len(pair_blob) != 33:
            raise ScriptError("uncompressed key in witness", errno.WITNESS_PUBKEYTYPE)
    try:
        public_pair = sec_to_public_pair(pair_blob, generator, strict=verify_strict)
    except (ValueError, EncodingError):
        return False

    if signature_type not in sighash_cache:
        sighash_cache[signature_type] = vm.signature_for_hash_type_f(signature_type, blobs_to_delete, vm)

    try:
        if generator.verify(public_pair, sighash_cache[signature_type], sig_pair):
            return True
    except ValueError:
        pass
    return False


def checksigs(vm, sig_blobs, public_pair_blobs):
    sig_blobs_remaining = list(sig_blobs)
    flags = vm.flags
    sighash_cache = {}
    verify_witness_pubkeytype = flags & VERIFY_WITNESS_PUBKEYTYPE
    verify_strict = not not (flags & VERIFY_STRICTENC)
    any_nonblank = (flags & VERIFY_NULLFAIL) and any(len(s) > 0 for s in sig_blobs)

    while len(sig_blobs_remaining) > 0:
        sig_blob = sig_blobs_remaining.pop()
        try:
            sig_pair, signature_type = parse_and_check_signature_blob(sig_blob, flags, vm)
        except (der.UnexpectedDER, ValueError):
            public_pair_blobs = []
        while len(sig_blobs_remaining) < len(public_pair_blobs):
            pair_blob = public_pair_blobs.pop()
            if checksig(vm, sig_pair, signature_type, pair_blob, sig_blobs,
                        sighash_cache, verify_witness_pubkeytype, verify_strict):
                break
        else:
            if any_nonblank:
                raise ScriptError("bad signature not NULL", errno.NULLFAIL)
            vm.append(vm.VM_FALSE)
            return
    vm.append(vm.VM_TRUE)


def do_OP_CHECKSIG(vm):
    pair_blob = vm.pop()
    sig_blob = vm.pop()
    checksigs(vm, [sig_blob], [pair_blob])


def do_OP_CHECKMULTISIG(vm):
    key_count = vm.pop_int()
    if key_count < 0 or key_count > 20:
        raise ScriptError("key_count not in range 0 to 20", errno.PUBKEY_COUNT)
    public_pair_blobs = [vm.pop() for _ in range(key_count)]
    public_pair_blobs.reverse()

    signature_count = vm.pop_int()
    if signature_count < 0 or signature_count > key_count:
        raise ScriptError(
            "invalid number of signatures: %d for %d keys" % (signature_count, key_count), errno.SIG_COUNT)
    sig_blobs = [vm.pop() for _ in range(signature_count)]
    sig_blobs.reverse()

    # check that we have the required hack 00 byte
    hack_byte = vm.pop()
    if vm.flags & VERIFY_NULLDUMMY and hack_byte != b'':
        raise ScriptError("bad dummy byte in checkmultisig", errno.SIG_NULLDUMMY)

    checksigs(vm, sig_blobs, public_pair_blobs)

    vm.op_count += key_count


def do_OP_CHECKMULTISIGVERIFY(vm):
    do_OP_CHECKMULTISIG(vm)
    v = vm.bool_from_script_bytes(vm.pop())
    if not v:
        raise ScriptError("VERIFY failed", errno.VERIFY)


def do_OP_CHECKSIGVERIFY(vm):
    do_OP_CHECKSIG(vm)
    v = vm.bool_from_script_bytes(vm.pop())
    if not v:
        raise ScriptError("VERIFY failed", errno.VERIFY)


"""
Parse, stream, create, sign and verify Bitcoin transactions as Tx structures.


The MIT License (MIT)

Copyright (c) 2015 by Richard Kiss

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""




from ..SolutionChecker import ScriptError

from pycoin.encoding.hexbytes import b2h, h2b
from pycoin.satoshi.flags import SIGHASH_ALL
from pycoin.solve.constraints import Atom, Operator, make_traceback_f
from pycoin.solve.ConstraintSolver import SolvingError, ConstraintSolver
from pycoin.solve.some_solvers import register_all


def generate_default_placeholder_signature(generator):
    return h2b(
        "3045022100fffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd036414002207"
        "fffffffffffffffffffffffffffffff5d576e7357a4501ddfe92f46681b20a001")


class DynamicStack(list):
    def __init__(self, initial_stack=[], reserve_count=0, fill_template="x_%d"):
        self.total_item_count = reserve_count
        self.fill_template = fill_template
        super(DynamicStack, self).__init__(initial_stack)

    def _fill(self):
        self.insert(0, Atom(self.fill_template % self.total_item_count))
        self.total_item_count += 1

    def pop(self, i=-1):
        while len(self) < abs(i):
            self._fill()
        return super(DynamicStack, self).pop(i)

    def __getitem__(self, *args, **kwargs):
        while True:
            try:
                return super(DynamicStack, self).__getitem__(*args, **kwargs)
            except IndexError:
                self._fill()


class Solver(object):
    SolutionChecker = None
    ScriptTools = None

    def __init__(self, tx):
        self.tx = tx
        self.solution_checker = self.SolutionChecker(tx)
        # self.sighash_cache = {}

    def determine_constraints(self, tx_in_idx, p2sh_lookup={}):
        tx_context = self.solution_checker.tx_context_for_idx(tx_in_idx)
        tx_context.witness_solution_stack = DynamicStack([Atom("w_%d" % (1-_)) for _ in range(2)], fill_template="w_%d")
        script_hash = self.solution_checker.script_hash_from_script(tx_context.puzzle_script)
        witness_version = self.solution_checker._witness_program_version(tx_context.puzzle_script)
        tx_context.solution_script = b''
        solution_reserve_count = 0
        fill_template = "x_%d"
        if script_hash:
            underlying_script = p2sh_lookup.get(script_hash, None)
            if underlying_script is None:
                raise ValueError("p2sh_lookup not set or does not have script hash for %s" % b2h(script_hash))
            tx_context.solution_script = self.ScriptTools.compile_push_data_list([underlying_script])
            solution_reserve_count = 1
            witness_version = self.solution_checker._witness_program_version(underlying_script)
        if witness_version == 0:
            witness_program = (underlying_script if script_hash else tx_context.puzzle_script)[2:]
            if len(witness_program) == 32:
                underlying_script_wit = p2sh_lookup.get(witness_program, None)
                if underlying_script_wit is None:
                    raise ValueError("p2sh_lookup not set or does not have script hash for %s" % b2h(witness_program))
                fill_template = "w_%d"
                solution_reserve_count = 1
                tx_context.witness_solution_stack = [underlying_script_wit]
        constraints = []

        def reset_stack_f(stack):
            return DynamicStack(stack, solution_reserve_count, fill_template)

        try:
            traceback_f = make_traceback_f(constraints, self.ScriptTools.int_for_opcode, reset_stack_f)
            self.solution_checker.check_solution(tx_context, traceback_f=traceback_f)
        except ScriptError:
            pass
        if script_hash:
            constraints.append(Operator('EQUAL', Atom("x_0"), underlying_script))
        if witness_version == 0:
            if len(witness_program) == 32:
                constraints.append(Operator('EQUAL', Atom("w_0"), underlying_script_wit))
        return constraints

    def solve_for_constraints(self, constraints, **kwargs):
        solutions = []
        for c in constraints:
            s = self.solutions_for_constraint(c)
            # s = (solution_f, target atom, dependency atom list)
            if s:
                solutions.append(s)
        deps = set()
        for c in constraints:
            deps.update(c.dependencies())
        solved_values = {d: None for d in deps}
        progress = True
        while progress and None in solved_values.values():
            progress = False
            for solution, target, dependencies in solutions:
                if any(solved_values.get(t) is not None for t in target):
                    continue
                if any(solved_values[d] is None for d in dependencies):
                    continue
                s = solution(solved_values, **kwargs)
                solved_values.update(s)
                progress = progress or (len(s) > 0)

        x_keys = sorted((k for k in solved_values.keys() if k.name.startswith("x")), reverse=True)
        w_keys = sorted((k for k in solved_values.keys() if k.name.startswith("w")), reverse=True)
        solution_list = [solved_values.get(k) for k in x_keys]
        witness_list = [solved_values.get(k) for k in w_keys]
        return solution_list, witness_list

    def solve(self, hash160_lookup, tx_in_idx, hash_type=None, **kwargs):
        """
        Sign a standard transaction.
        hash160_lookup:
            An object with a get method that accepts a hash160 and returns the
            corresponding (secret exponent, public_pair, is_compressed) tuple or
            None if it's unknown (in which case the script will obviously not be signed).
            A standard dictionary will do nicely here.
        tx_in_idx:
            the index of the tx_in we are currently signing
        """
        if hash_type is None:
            hash_type = SIGHASH_ALL
        kwargs["hash160_lookup"] = hash160_lookup
        if "signature_placeholder" not in kwargs:
            kwargs["signature_placeholder"] = generate_default_placeholder_signature(kwargs.get("generator"))
        if self.tx.txs_in[tx_in_idx].witness:
            kwargs["existing_script"] = self.tx.txs_in[tx_in_idx].witness
        else:
            kwargs["existing_script"] = [
                data for opcode, data, pc, new_pc in self.ScriptTools.get_opcodes(
                    self.tx.txs_in[tx_in_idx].script) if data is not None]
        kwargs["signature_type"] = hash_type
        kwargs["generator_for_signature_type_f"] = self.solution_checker.VM.generator_for_signature_type
        constraints = self.determine_constraints(tx_in_idx, p2sh_lookup=kwargs.get("p2sh_lookup"))
        solution_list, witness_list = self.solve_for_constraints(constraints, **kwargs)
        solution_script = self.ScriptTools.compile_push_data_list(solution_list)
        if witness_list:
            return solution_script, witness_list
        return solution_script

    def sign(self, hash160_lookup, tx_in_idx_set=None, hash_type=None, **kwargs):
        """
        Sign a standard transaction.
        hash160_lookup:
            A dictionary (or another object with .get) where keys are hash160 and
            values are tuples (secret exponent, public_pair, is_compressed) or None
            (in which case the script will obviously not be signed).
        """
        checker = self.SolutionChecker(self.tx)
        if tx_in_idx_set is None:
            tx_in_idx_set = range(len(self.tx.txs_in))
        self.tx.check_unspents()
        for tx_in_idx in sorted(tx_in_idx_set):
            tx_context = checker.tx_context_for_idx(tx_in_idx)
            try:
                checker.check_solution(tx_context, flags=None)
                continue
            except ScriptError:
                pass
            try:
                r = self.solve(hash160_lookup, tx_in_idx, hash_type=hash_type, **kwargs)
                if isinstance(r, bytes):
                    self.tx.txs_in[tx_in_idx].script = r
                else:
                    self.tx.txs_in[tx_in_idx].script = r[0]
                    self.tx.set_witness(tx_in_idx, r[1])
            except (SolvingError, ValueError):
                pass
        return self


BitcoinConstraintSolver = ConstraintSolver()
register_all(BitcoinConstraintSolver)

from .ScriptTools import BitcoinScriptTools
from .SolutionChecker import BitcoinSolutionChecker


class BitcoinSolver(Solver):
    SolutionChecker = BitcoinSolutionChecker
    ScriptTools = BitcoinScriptTools

    def solutions_for_constraint(self, c):
        return BitcoinConstraintSolver.solutions_for_constraint(c)



from pycoin.satoshi import opcodes
from pycoin.satoshi.IntStreamer import IntStreamer
from pycoin.vm.ScriptTools import ScriptTools

from .ScriptStreamer import BitcoinScriptStreamer

BitcoinScriptTools = ScriptTools(opcodes.OPCODE_LIST, IntStreamer, BitcoinScriptStreamer)



from pycoin.networks.bitcoinish import create_bitcoinish_network

network = create_bitcoinish_network(
    symbol="STRAT", network_name="Strat", subnet_name="mainnet",
    wif_prefix_hex="bf", sec_prefix="STRATSEC:", address_prefix_hex="3f", pay_to_script_prefix_hex="7d",
    bip32_prv_prefix_hex="0488ADE4", bip32_pub_prefix_hex="0488B21E",
    default_port=16178)



#!/usr/bin/env python

from __future__ import print_function

import argparse
import sqlite3
import sys

from pycoin.key.subpaths import subpaths_for_path_range
from pycoin.networks.registry import (
    network_codes, network_for_netcode
)


def create_parser():
    codes = network_codes()
    parser = argparse.ArgumentParser(
        description=(
            'Cache look-up information into a Keychain for use with tx. '
            'Useful for hiearchical keys with many children.'),
        epilog=('Known networks codes:\n  ' +
                ', '.join(['%s (%s)' % (i, network_for_netcode(i).full_name()) for i in codes]))
    )
    parser.add_argument('-n', "--netcode", help='specify network by netcode', choices=codes, default="BTC")
    parser.add_argument('-m', "--multisig", metavar="sigcount", type=int,
                        help='multisig, with this many signatures need to unencumber the funds')
    parser.add_argument('keychain', help='the keychain file (SQLite3 formatted)')
    parser.add_argument('subkey_paths', help='subkey paths (example: 0H/2/15-20)')
    parser.add_argument('key', nargs="+", help='a hierarchical wallet key string (public suffices)')
    return parser


def keychain(args, parser):
    network = network_for_netcode(args.netcode)

    parse = network.ui.parse

    keychain = network.keychain(sqlite3.connect(args.keychain))

    keys = []
    for _ in args.key:
        key = parse(_, types=["electrum", "bip32"])
        if not key:
            raise ValueError("can't parse %s" % _)
        keys.append(key)

    subkey_paths = args.subkey_paths

    m = args.multisig
    if m and m > len(keys):
        raise ValueError("not enough keys for %d signatures" % m)

    total_paths = 0

    for path in subpaths_for_path_range(subkey_paths):
        if m:
            secs = sorted([_.subkey_for_path(path).sec() for _ in keys])
            script = network.contract.for_multisig(m, secs)
            keychain.add_p2s_script(script)
            print(network.ui.address_for_p2s(script))
        total_paths += keychain.add_keys_path(keys, path)
    keychain.commit()
    print("%d total paths" % total_paths, file=sys.stderr)


def main():
    parser = create_parser()
    args = parser.parse_args()
    keychain(args, parser)


if __name__ == '__main__':
    main()



import io
import json

from .agent import urlopen

from pycoin.coins.bitcoin.Tx import Tx
from pycoin.encoding.hexbytes import b2h_rev, h2b, h2b_rev
from pycoin.networks.default import get_current_netcode


class ChainSoProvider(object):
    def __init__(self, netcode=None):
        NETWORK_PATHS = {
            "BTC": "BTC",
            "XTN": "BTCTEST",
            "DOGE": "DOGE",
            "XDT": "DOGETEST",
        }
        if netcode is None:
            netcode = get_current_netcode()
        self.network_path = NETWORK_PATHS.get(netcode)

    def base_url(self, method, args):
        return "https://chain.so/api/v2/%s/%s/%s" % (method, self.network_path, args)

    def spendables_for_address(self, address):
        """
        Return a list of Spendable objects for the
        given bitcoin address.
        """
        spendables = []
        r = json.loads(urlopen(self.base_url('get_tx_unspent', address)).read().decode("utf8"))

        for u in r['data']['txs']:
            coin_value = int(float(u['value']) * 100000000)
            script = h2b(u["script_hex"])
            previous_hash = h2b_rev(u["txid"])
            previous_index = u["output_no"]
            spendables.append(Tx.Spendable(coin_value, script, previous_hash, previous_index))

        return spendables

    def tx_for_tx_hash(self, tx_hash):
        "Get a Tx by its hash."
        url = self.base_url("get_tx", b2h_rev(tx_hash))
        r = json.loads(urlopen(url).read().decode("utf8"))
        tx = Tx.parse(io.BytesIO(h2b(r.get("data").get("tx_hex"))))
        return tx



import unittest

from pycoin.symbols.btc import network


class ParseTest(unittest.TestCase):

    def test_parse_wif(self):
        WIF = "KwDiBf89QgGbjEhKnhXJuH7LrciVrZi3qYjgd9M7rFU73sVHnoWn"
        k = network.parse.wif(WIF)
        self.assertEqual(k.secret_exponent(), 1)

    def test_parse_bip32_prv(self):
        HWIF = (
            "xprv9s21ZrQH143K31AgNK5pyVvW23gHnkBq2wh5aEk6g1s496M8ZMj"
            "xncCKZKgb5jZoY5eSJMJ2Vbyvi2hbmQnCuHBujZ2WXGTux1X2k9Krdtq")
        k = network.parse.bip32_prv(HWIF)
        self.assertEqual(k.as_text(as_private=True), HWIF)

    def test_parse_bip32_pub(self):
        HWIF = (
            "xpub661MyMwAqRbcFVF9ULcqLdsEa5WnCCugQAcgNd9iEMQ31tgH6u4DL"
            "QWoQayvtSVYFvXz2vPPpbXE1qpjoUFidhjFj82pVShWu9curWmb2zy")
        k = network.parse.bip32_pub(HWIF)
        self.assertEqual(k.as_text(), HWIF)

    def test_parse_bip32_seed(self):
        expected_hwif = (
            "xprv9s21ZrQH143K31AgNK5pyVvW23gHnkBq2wh5aEk6g1s496M8ZMjxn"
            "cCKZKgb5jZoY5eSJMJ2Vbyvi2hbmQnCuHBujZ2WXGTux1X2k9Krdtq")
        k = network.parse.bip32_seed("P:foo")
        self.assertEqual(k.as_text(as_private=True), expected_hwif)
        k = network.parse.bip32_seed("H:666f6f")
        self.assertEqual(k.as_text(as_private=True), expected_hwif)

    def test_parse_electrum_seed(self):
        E_INIT = "E:00000000000000000000000000000001"
        k = network.parse.electrum_seed(E_INIT)
        self.assertEqual(
            k.secret_exponent(),
            0x2ccdb632d4630c8e5a417858f70876afe5585c15b1c0940771af9ac160201b1d)

    def test_parse_electrum_prv(self):
        E_PRV = "E:0000000000000000000000000000000000000000000000000000000000000001"
        k = network.parse.electrum_prv(E_PRV)
        self.assertEqual(k.secret_exponent(), 1)

    def test_parse_electrum_pub(self):
        E_PUB = (
            "E:"
            "79be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798"
            "483ada7726a3c4655da4fbfc0e1108a8fd17b448a68554199c47d08ffb10d4b8")
        k = network.parse.electrum_pub(E_PUB)
        self.assertEqual(k.address(), "1EHNa6Q4Jz2uvNExL497mE43ikXhwF6kZm")

    def test_parse_address_p2pkh(self):
        address = "1BgGZ9tcN4rm9KBzDn7KprQz87SZ26SAMH"
        k = network.parse.p2pkh(address)
        self.assertEqual(k.address(), address)

    def test_parse_address_p2sh(self):
        address = "3JvL6Ymt8MVWiCNHC7oWU6nLeHNJKLZGLN"
        k = network.parse.p2sh(address)
        self.assertEqual(k.address(), address)

    def test_parse_address_p2pkh_wit(self):
        address = "bc1qw508d6qejxtdg4y5r3zarvary0c5xw7kv8f3t4"
        k = network.parse.p2pkh_segwit(address)
        self.assertEqual(k.address(), address)

    def test_parse_address_p2sh_wit(self):
        import hashlib
        address = network.address.for_p2sh_wit(hashlib.sha256(b'').digest())

        k = network.parse.p2sh_segwit(address)
        self.assertEqual(k.address(), address)

    def test_parse_script(self):
        k = network.parse.script("foo")
        self.assertEqual(k, None)
        k = network.parse.script("OP_CHECKSIG")
        self.assertEqual(k.disassemble(), "OP_CHECKSIG")



import hashlib

from pycoin.contrib import segwit_addr
from pycoin.encoding.b58 import b2a_hashed_base58
from pycoin.encoding.hash import hash160
from pycoin.encoding.hexbytes import b2h
from pycoin.intbytes import iterbytes


def make_address_api(
        contracts,
        bip32_prv_prefix=None, bip32_pub_prefix=None, wif_prefix=None,
        sec_prefix=None, address_prefix=None, pay_to_script_prefix=None,
        bech32_hrp=None):

    class AddressAPI(object):

        def for_script(self, script):
            info = contracts.info_for_script(script)
            return self.for_script_info(info)

        def b2a(self, blob):
            return b2a_hashed_base58(blob)

        # address_for_script and script_for_address stuff
        def for_script_info(self, script_info):
            type = script_info.get("type")

            if type == "p2pkh":
                return self.for_p2pkh(script_info["hash160"])

            if type == "p2pkh_wit":
                return self.for_p2pkh_wit(script_info["hash160"])

            if type == "p2sh_wit":
                return self.for_p2sh_wit(script_info["hash256"])

            if type == "p2pk":
                h160 = hash160(script_info["sec"])
                # BRAIN DAMAGE: this isn't really a p2pkh
                return self.for_p2pkh(h160)

            if type == "p2sh":
                return self.for_p2sh(script_info["hash160"])

            if type == "nulldata":
                return "(nulldata %s)" % b2h(script_info["data"])

            return "???"

        if address_prefix:
            def for_p2pkh(self, h160):
                return self.b2a(address_prefix + h160)

        if pay_to_script_prefix:
            def for_p2sh(self, h160):
                return self.b2a(pay_to_script_prefix + h160)

        if bech32_hrp:
            def for_p2pkh_wit(self, h160):
                assert len(h160) == 20
                return segwit_addr.encode(bech32_hrp, 0, iterbytes(h160))

        if bech32_hrp:
            def for_p2sh_wit(self, hash256):
                assert len(hash256) == 32
                return segwit_addr.encode(bech32_hrp, 0, iterbytes(hash256))

        # p2s and p2s_wit helpers

        if pay_to_script_prefix:
            def for_p2s(self, script):
                return self.for_p2sh(hash160(script))

        if bech32_hrp:
            def for_p2s_wit(self, script):
                return self.for_p2sh_wit(hashlib.sha256(
                    script).digest())

    return AddressAPI()



from pycoin.networks.bitcoinish import create_bitcoinish_network

network = create_bitcoinish_network(
    symbol="TPIVX", network_name="PIVX", subnet_name="testnet",
    wif_prefix_hex="ef", address_prefix_hex="8b", pay_to_script_prefix_hex="13",
    bip32_prv_prefix_hex="3a8061a0", bip32_pub_prefix_hex="3a805837")



from ..intbytes import int2byte

from .base_conversion import EncodingError
from .bytes32 import from_bytes_32, to_bytes_32
from .hash import hash160


def public_pair_to_sec(public_pair, compressed=True):
    """Convert a public pair (a pair of bignums corresponding to a public key) to the
    gross internal sec binary format used by OpenSSL."""
    x_str = to_bytes_32(public_pair[0])
    if compressed:
        return int2byte((2 + (public_pair[1] & 1))) + x_str
    y_str = to_bytes_32(public_pair[1])
    return b'\4' + x_str + y_str


def sec_to_public_pair(sec, generator=None, strict=True):
    """Convert a public key in sec binary format to a public pair."""
    byte_count = (generator.p().bit_length() + 7) >> 3 if generator else (len(sec) - 1)
    x = from_bytes_32(sec[1:1 + byte_count])
    sec0 = sec[:1]
    if len(sec) == 1 + byte_count * 2:
        isok = sec0 == b'\4'
        if not strict:
            isok = isok or (sec0 in [b'\6', b'\7'])
        if isok:
            y = from_bytes_32(sec[1+byte_count:1+2*byte_count])
            return (x, y)
    elif len(sec) == 1 + byte_count:
        if not strict or (sec0 in (b'\2', b'\3')):
            is_y_odd = (sec0 != b'\2')
            return generator.points_for_x(x)[is_y_odd]
    raise EncodingError("bad sec encoding for public key")


def is_sec(sec):
    c = sec[:1]
    size = len(sec)
    if c in (b'\2', b'\3') and size == 33:
        return True
    return c == b'\4' and size == 65


def is_sec_compressed(sec):
    """Return a boolean indicating if the sec represents a compressed public key."""
    return sec[:1] in (b'\2', b'\3')


def public_pair_to_hash160_sec(public_pair, compressed=True):
    """Convert a public_pair (corresponding to a public key) to hash160_sec format.
    This is a hash of the sec representation of a public key, and is used to generate
    the corresponding Bitcoin address."""
    return hash160(public_pair_to_sec(public_pair, compressed=compressed))


"""
The MIT License (MIT)

Copyright (c) 2013 by Richard Kiss

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""



OK = 0
UNKNOWN_ERROR = 1
EVAL_FALSE = 2
OP_RETURN = 3

# Max sizes
SCRIPT_SIZE = 4
PUSH_SIZE = 5
OP_COUNT = 6
STACK_SIZE = 7
SIG_COUNT = 8
PUBKEY_COUNT = 9

# Failed verify operations
VERIFY = 10
EQUALVERIFY = 11
CHECKMULTISIGVERIFY = 12
CHECKSIGVERIFY = 13
NUMEQUALVERIFY = 14

# Logical/Format/Canonical errors
BAD_OPCODE = 15
DISABLED_OPCODE = 16
INVALID_STACK_OPERATION = 17
INVALID_ALTSTACK_OPERATION = 18
UNBALANCED_CONDITIONAL = 19

# CHECKLOCKTIMEVERIFY and CHECKSEQUENCEVERIFY
NEGATIVE_LOCKTIME = 20
UNSATISFIED_LOCKTIME = 21

# Malleability
SIG_HASHTYPE = 22
SIG_DER = 23
MINIMALDATA = 24
SIG_PUSHONLY = 25
SIG_HIGH_S = 26
SIG_NULLDUMMY = 27
PUBKEYTYPE = 28
CLEANSTACK = 29
MINIMALIF = 30
NULLFAIL = 31

# softfork safeness
DISCOURAGE_UPGRADABLE_NOPS = 32
DISCOURAGE_UPGRADABLE_WITNESS_PROGRAM = 33

# segregated witness
WITNESS_PROGRAM_WRONG_LENGTH = 34
WITNESS_PROGRAM_WITNESS_EMPTY = 35
WITNESS_PROGRAM_MISMATCH = 36
WITNESS_MALLEATED = 37
WITNESS_MALLEATED_P2SH = 38
WITNESS_UNEXPECTED = 39
WITNESS_PUBKEYTYPE = 40

ERROR_COUNT = 41



import unittest

from pycoin.encoding.hexbytes import h2b, h2b_rev, b2h, b2h_rev


class HexbytesTest(unittest.TestCase):

    def test_h2b(self):
        h = "000102"
        b = b"\x00\x01\x02"
        self.assertEqual(h2b(h), b)
        self.assertEqual(b2h(b), h)
        self.assertEqual(h2b_rev(h), b[::-1])
        self.assertEqual(b2h_rev(b), "020100")


if __name__ == '__main__':
    unittest.main()



import unittest
import json
import os

from pycoin.encoding.hexbytes import h2b_rev
from pycoin.symbols.btc import network


flags = network.validator.flags
ValidationFailureError = network.validator.ValidationFailureError

DEBUG_TX_ID_LIST = []


TX_VALID_JSON = os.path.dirname(__file__) + '/data/tx_valid.json'
TX_INVALID_JSON = os.path.dirname(__file__) + '/data/tx_invalid.json'


def parse_flags(flag_string):
    v = 0
    if len(flag_string) > 0:
        for f in flag_string.split(","):
            v |= getattr(flags, "VERIFY_%s" % f)
    return v


def txs_from_json(path):
    """
    Read tests from ./data/tx_??valid.json
    Format is an array of arrays
    Inner arrays are either [ "comment" ]
    or [[[prevout hash, prevout index, prevout scriptPubKey], [input 2], ...], serializedTransaction, verifyFlags]
    ... where all scripts are stringified scripts.

    verifyFlags is a comma separated list of script verification flags to apply, or "NONE"
    """
    comments = None
    with open(path, 'r') as f:
        for tvec in json.load(f):
            if len(tvec) == 1:
                comments = tvec[0]
                continue
            assert len(tvec) == 3
            prevouts = tvec[0]
            for prevout in prevouts:
                assert len(prevout) in (3, 4)

            tx_hex = tvec[1]

            flag_mask = parse_flags(tvec[2])
            try:
                tx = network.tx.from_hex(tx_hex)
            except Exception:
                print("Cannot parse tx_hex: %s" % tx_hex)
                raise

            spendable_db = {}
            blank_spendable = network.tx.Spendable(0, b'', b'\0' * 32, 0)
            for prevout in prevouts:
                coin_value = 1000000
                if len(prevout) == 4:
                    coin_value = prevout[3]
                spendable = network.tx.Spendable(
                    coin_value=coin_value, script=network.script.compile(prevout[2]),
                    tx_hash=h2b_rev(prevout[0]), tx_out_index=prevout[1])
                spendable_db[(spendable.tx_hash, spendable.tx_out_index)] = spendable
            unspents = [
                spendable_db.get((tx_in.previous_hash, tx_in.previous_index), blank_spendable) for tx_in in tx.txs_in]
            tx.set_unspents(unspents)
            yield (tx, flag_mask, comments)


class TestTx(unittest.TestCase):
    pass


def make_f(tx, flag_mask, comments, expect_ok=True):
    tx_hex = tx.as_hex(include_unspents=True)

    def test_f(self):
        why = None
        try:
            tx.check()
        except ValidationFailureError as ex:
            why = str(ex)
        bs = 0
        for tx_in_idx in range(len(tx.txs_in)):
            try:
                if DEBUG_TX_ID_LIST:
                    import pdb
                    pdb.set_trace()
                tx.check_solution(tx_in_idx=tx_in_idx, flags=flag_mask)
            except tx.SolutionChecker.ScriptError as se:
                bs += 1
        if bs > 0:
            why = "bad sig count = %d" % bs
        if (why is not None) == expect_ok:
            why = why or "tx unexpectedly validated"
            f = open("tx-%s-%x-%s.hex" % (tx.w_id(), flag_mask, "valid" if expect_ok else "invalid"), "w")
            f.write(tx_hex)
            f.close()
            self.fail("fail on %s because of %s with hex %s: %s" % (tx.w_id(), why, tx_hex, comments))
    if DEBUG_TX_ID_LIST and tx.w_id() not in DEBUG_TX_ID_LIST:
        return lambda self: 0
    return test_f


def inject():
    for idx, (tx, flag_mask, comments) in enumerate(txs_from_json(TX_VALID_JSON)):
        name_of_f = "test_valid_%02d_%s" % (idx, tx.w_id())
        setattr(TestTx, name_of_f, make_f(tx, flag_mask, comments))
        print("adding %s" % name_of_f)

    for idx, (tx, flag_mask, comments) in enumerate(txs_from_json(TX_INVALID_JSON)):
        name_of_f = "test_invalid_%02d_%s" % (idx, tx.w_id())
        setattr(TestTx, name_of_f, make_f(tx, flag_mask, comments, expect_ok=False))
        print("adding %s" % name_of_f)


inject()


if __name__ == '__main__':
    unittest.main()


"""
Test suite for pycoin library: check validity of txs in files
tx_valid.json and tx_invalid.json. Adapted from Bitcoin Core
transaction_tests.cpp test suite.

The MIT License (MIT)

Copyright (c) 2015 by Marek Miller
Copyright (c) 2015 by Richard Kiss
Copyright (c) 2015 by The Bitcoin Core Developers

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""



from pycoin.encoding.b58 import a2b_base58
from pycoin.encoding.hash import double_sha256
from pycoin.contrib import segwit_addr


class parseable_str(str):
    """
    This is a subclass of str which allows caching of parsed base58 and bech32
    data (or really anything) to eliminate the need to repeatedly run slow parsing
    code when checking validity for multiple types.
    """
    def __new__(self, s):
        if isinstance(s, parseable_str):
            return s
        return str.__new__(self, s)

    def __init__(self, s):
        super(str, self).__init__()
        if isinstance(s, parseable_str):
            self._cache = s._cache
        else:
            self._cache = {}

    def cache(self, key, f):
        if key not in self._cache:
            self._cache[key] = None
            try:
                self._cache[key] = f(self)
            except Exception:
                pass
        return self._cache[key]


def parse_b58(s):
    s = parseable_str(s)
    return s.cache("b58", a2b_base58)


def b58_double_sha256(s):
    data = parse_b58(s)
    if data:
        data, the_hash = data[:-4], data[-4:]
        if double_sha256(data)[:4] == the_hash:
            return data


def parse_b58_double_sha256(s):
    s = parseable_str(s)
    return s.cache("b58_double_sha256", b58_double_sha256)


def parse_bech32(s):
    s = parseable_str(s)
    return s.cache("bech32", segwit_addr.bech32_decode)


def parse_colon_prefix(s):
    s = parseable_str(s)
    return s.cache("colon_prefix", lambda _: _.split(":", 1))



from pycoin.networks.bitcoinish import create_bitcoinish_network

network = create_bitcoinish_network(
    symbol="BTX", network_name="BitCore", subnet_name="mainnet",
    wif_prefix_hex="80", sec_prefix="BTXSEC:", address_prefix_hex="03", pay_to_script_prefix_hex="7D",
    bip32_prv_prefix_hex="0488ADE4", bip32_pub_prefix_hex="0488B21E",
    magic_header_hex="F9BEB4D9", default_port=8555,
    dns_bootstrap=[
        "seed.bitcore.biz"
    ])




import io

from pycoin.encoding.hash import double_sha256
from pycoin.satoshi.satoshi_struct import parse_struct, stream_struct

from pycoin.block import Block as BaseBlock

from .Tx import Tx


# Allow NON_FORKID in legacy tests and blocks under BTG hard fork height
ALLOW_NON_FORKID = (1 << 17)

# BRAIN DAMAGE: TODO: implement this in vm


class Block(BaseBlock):
    """A Block is an element of the Bitcoin chain."""

    Tx = Tx

    FORK_BLOCK = 491407

    @classmethod
    def parse_as_header(class_, f):
        """
        Parse the Block header from the file-like object
        """
        version, previous_block_hash, merkle_root, height = parse_struct("L##L", f)
        # https://github.com/BTCGPU/BTCGPU/wiki/Technical-Spec
        f.read(28)  # reserved area
        (timestamp, difficulty, nonce, solution) = parse_struct("LL#S", f)
        return class_(version, previous_block_hash, merkle_root, timestamp,
                      difficulty, nonce, height, solution)

    def __init__(self, version, previous_block_hash, merkle_root, timestamp,
                 difficulty, nonce, height, solution):
        self.version = version
        self.previous_block_hash = previous_block_hash
        self.merkle_root = merkle_root
        self.timestamp = timestamp
        self.difficulty = difficulty
        self.nonce = nonce
        self.height = height
        self.solution = solution
        self.txs = []

    def _calculate_hash(self):
        s = io.BytesIO()
        if self.height < self.FORK_BLOCK:
            self.stream_header_legacy(s)
        else:
            self.stream_header(s)
        return double_sha256(s.getvalue())

    def as_blockheader(self):
        return Block(self.version, self.previous_block_hash, self.merkle_root,
                     self.timestamp, self.difficulty, self.nonce,
                     self.height, self.solution)

    def stream_header_legacy(self, f):
        """Stream the block header in the standard way to the file-like object f."""
        stream_struct("L##LL", f, self.version, self.previous_block_hash,
                      self.merkle_root, self.timestamp, self.difficulty)
        f.write(self.nonce[:4])

    def stream_header(self, f):
        """Stream the block header in the standard way to the file-like object f."""
        stream_struct("L##L", f, self.version, self.previous_block_hash,
                      self.merkle_root, self.height)
        f.write(b'\0' * 28)
        stream_struct("LL#S", f, self.timestamp, self.difficulty, self.nonce, self.solution)



from pycoin.networks.bitcoinish import create_bitcoinish_network

from pycoin.coins.bgold.Tx import Tx as BgoldTx


network = create_bitcoinish_network(
    symbol="XTG", network_name="Bgold", subnet_name="testnet", tx=BgoldTx,
    wif_prefix_hex="ef", sec_prefix="XTNSEC:", address_prefix_hex="6f", pay_to_script_prefix_hex="c4",
    bip32_prv_prefix_hex="0488ade4", bip32_pub_prefix_hex="0488B21E", bech32_hrp="tb",
    magic_header_hex="e1476d44", default_port=18338,
    dns_bootstrap=[
        "eu-test-dnsseed.bitcoingold-official.org", "test-dnsseed.bitcoingold.org",
        "test-dnsseed.btcgpu.org", "btg.dnsseed.minertopia.org"
    ])



import io

from .SolutionChecker import SolutionChecker, ScriptError
from .TxIn import TxIn
from .TxOut import TxOut

from pycoin.encoding.hexbytes import b2h, b2h_rev, h2b


class Tx(object):
    TxIn = TxIn
    TxOut = TxOut
    Spendable = None
    Solver = None
    SolutionChecker = SolutionChecker

    @classmethod
    def parse(class_, f):
        """Parse a transaction Tx from the file-like object f."""
        raise NotImplemented()

    @classmethod
    def from_bin(class_, blob):
        """Return the Tx for the given binary blob.

        :param blob: a binary blob containing a transaction streamed in standard
            form. The blob may also include the unspents (a nonstandard extension,
            optionally written by :func:`Tx.stream <stream>`), and they will also be parsed.
        :return: :class:`Tx`

        If parsing fails, an exception is raised.
        """
        f = io.BytesIO(blob)
        tx = class_.parse(f)
        try:
            tx.parse_unspents(f)
        except Exception:
            # parsing unspents failed
            tx.unspents = []
        return tx

    @classmethod
    def from_hex(class_, hex_string):
        """Return the Tx for the given hex string.

        :param hex_string: a hex string containing a transaction streamed in standard
            form. The blob may also include the unspents (a nonstandard extension,
            optionally written by :func:`Tx.stream <stream>`), and they will also be parsed.
        :return: :class:`Tx`

        If parsing fails, an exception is raised.
        """
        return class_.from_bin(h2b(hex_string))

    def __init__(self, *args, **kwargs):
        raise NotImplemented()

    def stream(self, f, *args, **kwargs):
        """Stream a transaction Tx to the file-like object f."""
        raise NotImplemented()

    def as_bin(self, *args, **kwargs):
        """Returns a binary blob containing the streamed transaction.

        For information about the parameters, see :func:`Tx.stream <stream>`

        :return: binary blob that would parse to the given transaction
        """
        f = io.BytesIO()
        self.stream(f, *args, **kwargs)
        return f.getvalue()

    def as_hex(self, *args, **kwargs):
        """Returns a text string containing the streamed transaction encoded as hex.

        For information about the parameters, see :func:`Tx.stream <stream>`

        :return: hex string that would parse to the given transaction
        """
        return b2h(self.as_bin(*args, **kwargs))

    def hash(self, hash_type=None):
        """Return the hash for this Tx object."""
        raise NotImplemented()

    def id(self):
        """Return the human-readable hash for this Tx object."""
        return b2h_rev(self.hash())

    def total_out(self):
        return sum(tx_out.coin_value for tx_out in self.txs_out)

    def tx_outs_as_spendable(self, block_index_available=0):
        h = self.hash()
        return [
            self.Spendable.from_tx_out(tx_out, h, tx_out_index, block_index_available)
            for tx_out_index, tx_out in enumerate(self.txs_out)]

    def __str__(self):
        raise NotImplemented()

    def __repr__(self):
        raise NotImplemented()

    def check(self):
        """
        Basic checks that don't depend on network or block context.
        """
        raise NotImplemented()

    """
    The functions below here deal with an optional additional parameter: "unspents".
    This parameter is a list of tx_out objects that are referenced by the
    list of self.tx_in objects.
    """

    def set_unspents(self, unspents):
        """
        Set the unspent inputs for a transaction.

        :param unspents: a list of :class:`TxOut` (or the subclass :class:`Spendable`) objects
            corresponding to the :class:`TxIn` objects for this transaction (same number of
            items in each list)
        """
        if len(unspents) != len(self.txs_in):
            raise ValueError("wrong number of unspents")
        self.unspents = unspents

    def sign(self, *args, **kwargs):
        """
        Sign all transaction inputs. The parameters vary depending upon the way the coins being
        spent are encumbered.
        """
        self.Solver(self).sign(*args, **kwargs)
        return self

    def check_solution(self, tx_in_idx, *args, **kwargs):
        sc = self.SolutionChecker(self)
        tx_context = sc.tx_context_for_idx(tx_in_idx)
        sc.check_solution(tx_context, *args, **kwargs)

    def is_solution_ok(self, tx_in_idx, *args, **kwargs):
        if len(self.unspents) <= tx_in_idx or self.unspents[tx_in_idx] is None:
            return False
        try:
            self.check_solution(tx_in_idx, *args, **kwargs)
            return True
        except ScriptError:
            return False

    def bad_solution_count(self, *args, **kwargs):
        "Return a count of how many :class:`TxIn` objects are not correctly solved."
        return sum(0 if self.is_solution_ok(idx, *args, **kwargs) else 1 for idx in range(len(self.txs_in)))


"""
The MIT License (MIT)

Copyright (c) 2013 by Richard Kiss

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""



from pycoin.encoding.bytes32 import from_bytes_32, to_bytes_32
from pycoin.encoding.hash import hash160
from pycoin.encoding.hexbytes import b2h
from pycoin.encoding.sec import (
    is_sec_compressed, public_pair_to_sec, sec_to_public_pair
)
from pycoin.satoshi.der import sigencode_der, sigdecode_der


class InvalidPublicPairError(ValueError):
    pass


class InvalidSecretExponentError(ValueError):
    pass


class Key(object):

    _network = None
    _generator = None

    @classmethod
    def make_subclass(class_, network, generator):

        class Key(class_):
            _network = network
            _generator = generator

        return Key

    def __init__(self, secret_exponent=None, public_pair=None, is_compressed=True):
        """
        Include at most one of secret_exponent or public_pair.

        secret_exponent:
            a long representing the secret exponent
        public_pair:
            a tuple of long integers on the ecdsa curve
        is_compressed:
            Is this key in the compressed form? The uncompressed form is obsolete.
            Note that any function which produces output that depends on this can override
            this default value.
        """
        if [secret_exponent, public_pair].count(None) != 1:
            raise ValueError("exactly one of secret_exponent or public_pair must be passed.")
        self._secret_exponent = secret_exponent
        self._public_pair = public_pair
        self._is_compressed = is_compressed
        self._hash160_uncompressed = None
        self._hash160_compressed = None

        if self._secret_exponent is not None:
            if self._secret_exponent < 1 \
                    or self._secret_exponent >= self._generator.order():
                raise InvalidSecretExponentError()
            public_pair = self._secret_exponent * self._generator
            self._public_pair = public_pair

        if (None in self._public_pair) or (
               not self._generator.contains_point(*self._public_pair)):
            raise InvalidPublicPairError()

    @classmethod
    def from_sec(class_, sec):
        """
        Create a key from an sec bytestream (which is an encoding of a public pair).
        """
        public_pair = sec_to_public_pair(sec, class_._generator)
        return class_(public_pair=public_pair, is_compressed=is_sec_compressed(sec))

    def is_private(self):
        return self.secret_exponent() is not None

    def secret_exponent(self):
        """
        Return an integer representing the secret exponent (or None).
        """
        return self._secret_exponent

    def wif(self, is_compressed=None):
        """
        Return the WIF representation of this key, if available.
        """
        secret_exponent = self.secret_exponent()
        if secret_exponent is None:
            return None
        if is_compressed is None:
            is_compressed = self.is_compressed()
        blob = to_bytes_32(secret_exponent)
        if is_compressed:
            blob += b'\01'
        return self._network.wif_for_blob(blob)

    def public_pair(self):
        """
        Return a pair of integers representing the public key (or None).
        """
        return self._public_pair

    def sec(self, is_compressed=None):
        """
        Return the SEC representation of this key, if available.
        """
        if is_compressed is None:
            is_compressed = self.is_compressed()
        public_pair = self.public_pair()
        if public_pair is None:
            return None
        return public_pair_to_sec(public_pair, compressed=is_compressed)

    def sec_as_hex(self, is_compressed=None):
        """
        Return the SEC representation of this key as hex text.
        """
        sec = self.sec(is_compressed=is_compressed)
        return self._network.sec_text_for_blob(sec)

    def hash160(self, is_compressed=None):
        """
        Return the hash160 representation of this key, if available.
        """
        if is_compressed is None:
            is_compressed = self.is_compressed()
        if is_compressed:
            if self._hash160_compressed is None:
                self._hash160_compressed = hash160(self.sec(is_compressed=is_compressed))
            return self._hash160_compressed

        if self._hash160_uncompressed is None:
            self._hash160_uncompressed = hash160(self.sec(is_compressed=is_compressed))
        return self._hash160_uncompressed

    def fingerprint(self, is_compressed=None):
        return self.hash160(is_compressed=is_compressed)[:4]

    def address(self, is_compressed=None):
        """
        Return the public address representation of this key, if available.
        """
        return self._network.address.for_p2pkh(self.hash160(is_compressed=is_compressed))

    def as_text(self):
        """
        Return a textual representation of this key.
        """
        if self.secret_exponent():
            return self.wif()
        sec_hex = self.sec_as_hex()
        if sec_hex:
            return sec_hex
        return self.address()

    def public_copy(self):
        if self.secret_exponent() is None:
            return self
        return self.__class__(public_pair=self.public_pair(), is_compressed=self.is_compressed())

    def subkey_for_path(self, path):
        return self

    def subkey(self, path_to_subkey):
        """
        Return the Key corresponding to the hierarchical wallet's subkey
        """
        return self

    def subkeys(self, path_to_subkeys):
        """
        Return an iterator yielding Keys corresponding to the
        hierarchical wallet's subkey path (or just this key).
        """
        yield self

    def sign(self, h):
        """
        Return a der-encoded signature for a hash h.
        Will throw a RuntimeError if this key is not a private key
        """
        if not self.is_private():
            raise RuntimeError("Key must be private to be able to sign")
        val = from_bytes_32(h)
        r, s = self._generator.sign(self.secret_exponent(), val)
        return sigencode_der(r, s)

    def verify(self, h, sig):
        """
        Return whether a signature is valid for hash h using this key.
        """
        val = from_bytes_32(h)
        pubkey = self.public_pair()
        return self._generator.verify(pubkey, val, sigdecode_der(sig))

    def is_compressed(self):
        """
        Return whether this key has been marked as compressed when it was created.
        """
        return self._is_compressed

    def __repr__(self):
        r = self.public_copy()
        if r._network:
            s = r.as_text()
        elif r.sec():
            s = b2h(r.sec())
        else:
            s = b2h(r.hash160())
        if self.is_private():
            return "private_for <%s>" % s
        return "<%s>" % s




def crack_secret_exponent_from_k(generator, signed_value, sig, k):
    """
    Given a signature of a signed_value and a known k, return the secret exponent.
    """
    r, s = sig
    return ((s * k - signed_value) * generator.inverse(r)) % generator.order()


def crack_k_from_sigs(generator, sig1, val1, sig2, val2):
    """
    Given two signatures with the same secret exponent and K value, return that K value.
    """

    # s1 = v1 / k1 + (se * r1) / k1
    # s2 = v2 / k2 + (se * r2) / k2
    # and k = k1 = k2
    # so
    # k * s1 = v1 + (se * r1)
    # k * s2 = v2 + (se * r2)
    # so
    # k * s1 * r2 = r2 * v1 + (se * r1 * r2)
    # k * s2 * r1 = r1 * v2 + (se * r2 * r1)
    # so
    # k (s1 * r2 - s2 * r1) = r2 * v1 - r1 * v2
    # so
    # k = (r2 * v1 - r1 * v2) / (s1 * r2 - s2 * r1)

    r1, s1 = sig1
    r2, s2 = sig2
    if r1 != r2:
        raise ValueError("r values of signature do not match")
    k = (r2 * val1 - r1 * val2) * generator.inverse(r2 * s1 - r1 * s2)
    return k % generator.order()



import importlib
import os
import pkgutil


def search_prefixes():
    prefixes = ["pycoin.symbols"]
    try:
        prefixes = os.getenv("PYCOIN_NETWORK_PATHS", "").split() + prefixes
    except Exception:
        pass
    return prefixes


def network_for_netcode(symbol):
    symbol = symbol.upper()
    netcode = symbol.lower()
    for prefix in search_prefixes():
        try:
            module = importlib.import_module("%s.%s" % (prefix, netcode))
            if module.network.symbol.upper() == symbol:
                module.symbol = symbol
                return module.network
        except (AttributeError, ImportError):
            pass
    raise ValueError("no network with symbol %s found" % netcode)


def iterate_symbols():
    """
    Return an iterator yielding registered netcodes.
    """
    for prefix in search_prefixes():
        package = importlib.import_module(prefix)
        for importer, modname, ispkg in pkgutil.walk_packages(path=package.__path__, onerror=lambda x: None):
            network = network_for_netcode(modname)
            if network:
                yield network.symbol.upper()


def network_codes():
    return list(iterate_symbols())




import os.path

from pycoin.coins.bitcoin.Tx import Tx
from pycoin.encoding.hexbytes import b2h_rev


class TxDb(object):
    """
    This object can be used in many places that expect a dict.
    """
    def __init__(self, lookup_methods=[], read_only_paths=[], writable_cache_path=None):
        self.lookup_methods = lookup_methods
        self.read_only_paths = read_only_paths
        if writable_cache_path:
            self.read_only_paths.append(writable_cache_path)
        self.writable_cache_path = writable_cache_path
        if self.writable_cache_path and not os.path.exists(self.writable_cache_path):
            os.makedirs(self.writable_cache_path)

    def paths_for_hash(self, hash):
        name = b2h_rev(hash)
        for base_dir in self.read_only_paths:
            p = os.path.join(base_dir, "%s_tx.bin" % name)
            if os.path.exists(p):
                yield p

    def put(self, tx):
        name = b2h_rev(tx.hash())
        if self.writable_cache_path:
            try:
                path = os.path.join(self.writable_cache_path, "%s_tx.bin" % name)
                with open(path, "wb") as f:
                    tx.stream(f)
            except IOError:
                pass

    def get(self, key):
        for path in self.paths_for_hash(key):
            try:
                tx = Tx.parse(open(path, "rb"))
                if tx and tx.hash() == key:
                    return tx
            except IOError:
                pass
        for method in self.lookup_methods:
            try:
                tx = method(key)
                if tx and tx.hash() == key:
                    self.put(tx)
                    return tx
            except Exception:
                pass
        return None

    def __getitem__(self, key):
        raise NotImplemented

    def __setitem__(self, key, val):
        if val.hash() != key:
            raise ValueError("bad key %s for %s" % (b2h_rev(key), val))
        self.put(val)



from pycoin.networks.bitcoinish import create_bitcoinish_network


network = create_bitcoinish_network(
    network_name="Digibyte", symbol="DGB", subnet_name="mainnet",
    wif_prefix_hex="b0", sec_prefix="DGBSEC:", address_prefix_hex="1E", pay_to_script_prefix_hex="3F",
    bip32_prv_prefix_hex="0488ADE4", bip32_pub_prefix_hex="0488B21E", bech32_hrp="dgb")



import pytest
import unittest

from pycoin.encoding.bytes32 import to_bytes_32
from pycoin.encoding.hexbytes import h2b
from pycoin.symbols.grs import network as GroestlcoinMainnet


pytest.importorskip('groestlcoin_hash')


class GroestlcoinEncodingTestCase(unittest.TestCase):
    def test_p2pkh(self):
        def do_test(h160, address):
            self.assertEqual(GroestlcoinMainnet.address.for_p2pkh(h160), address)
            self.assertEqual(GroestlcoinMainnet.parse.p2pkh(address).address(), address)

        do_test(h2b('0000000000000000000000000000000000000000'), 'FVAiSujNZVgYSc27t6zUTWoKfAGxer42D4')

    def test_p2sh(self):
        def do_test(h160, redeem_script, address):
            self.assertEqual(GroestlcoinMainnet.contract.for_p2sh(h160), redeem_script)
            self.assertEqual(GroestlcoinMainnet.address.for_p2sh(h160), address)
            self.assertEqual(GroestlcoinMainnet.parse.p2sh(address).address(), address)

        do_test(
            h2b('2a84cf00d47f699ee7bbc1dea5ec1bdecb4ac154'),
            h2b('a9142a84cf00d47f699ee7bbc1dea5ec1bdecb4ac15487'), '35ZqQJcBQMZ1rsv8aSuJ2wkC7ohUFNJZ77')

    def test_wif(self):
        def do_test(sec_bytes, wif, address):
            parsed = GroestlcoinMainnet.parse.wif(wif)
            self.assertEqual(to_bytes_32(parsed.secret_exponent()), sec_bytes)
            self.assertEqual(parsed.wif(), wif)
            self.assertEqual(parsed.address(), address)

        do_test(
            h2b('0000000000000000000000000000000000000000000000000000000000000001'),
            '5HpHagT65TZzG1PH3CSu63k8DbpvD8s5ip4nEB3kEsreAmVmAPb', 'FiT6218RsUiTMyG5DA8bDjrNNuofYV2MdF')
        do_test(
            h2b('0000000000000000000000000000000000000000000000000000000000000001'),
            'KwDiBf89QgGbjEhKnhXJuH7LrciVrZi3qYjgd9M7rFU73sX6ptSt', 'Ffqz14cyvZYJavD76t6oHNDJnGiWcZMVxR')

    def test_bip32(self):
        xprv = ('xprvA41z7zogVVwxVSgdKUHDy1SKmdb533PjDz7J6N6mV6uS3ze1ai8FHa8kmH'
                'ScGpWmj4WggLyQjgPie1rFSruoUihUZREPSL39UNdE3GaoVXP')
        xpub = ('xpub6H1LXWLaKsWFhvm6RVpEL9P4KfRZSW7abD2ttkWP3SSQvnyA8FSVqNTEcY'
                'FgJS2UaFcxupHiYkro49S8yGasTvXEYBVPamhGW6cFJuY6NHg')
        parsed = GroestlcoinMainnet.parse.hierarchical_key(xprv)
        self.assertEqual(parsed.tree_depth(), 5)
        self.assertEqual(parsed.parent_fingerprint(), b'\xd8\x80\xd7\xd8')
        self.assertEqual(parsed.child_index(), 1000000000)
        self.assertEqual(parsed.chain_code(),
                         b'\xc7\x83\xe6{\x92\x1d+\xeb\x8fk8\x9c\xc6F\xd7&;AEp\x1d\xad\xd2\x16\x15H\xa8\xb0x\xe6^\x9e')
        self.assertEqual(parsed.secret_exponent(),
                         32162737660659799401901343156672072893797470137297259782459076395168682141640)
        self.assertEqual(parsed.hwif(as_private=True), xprv)
        self.assertEqual(parsed.hwif(as_private=False), xpub)

        self.assertEqual(parsed.wif(), 'Kybw8izYevo5xMh1TK7aUr7jHFCxXS1zv8p3oqFz3o2zFbjGCYKc')
        self.assertEqual(parsed.address(), 'FpjSHmYBamwxWSsPtaL2rUz98FSS17VmY2')
        self.assertEqual(parsed.wif(is_compressed=False), '5JMbvQZXHJAzJyoDnqWasGCwtiHJZivF2ckjn3n5mazYYwQHway')
        self.assertEqual(parsed.address(is_compressed=False), 'FrH6KqPfsLPHArqCJ8WciGyvnWcRmvAsHC')



import unittest

from pycoin.encoding.hexbytes import h2b
from pycoin.symbols.btc import network as BTC
from pycoin.symbols.xtn import network as XTN


class Bip0032TestCase(unittest.TestCase):

    def test_vector_1(self):
        master = BTC.keys.bip32_seed(h2b("000102030405060708090a0b0c0d0e0f"))
        self.assertEqual(
            master.hwif(as_private=True),
            "xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPG"
            "JxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi")
        self.assertEqual(master.address(), "15mKKb2eos1hWa6tisdPwwDC1a5J1y9nma")
        self.assertEqual(master.wif(), "L52XzL2cMkHxqxBXRyEpnPQZGUs3uKiL3R11XbAdHigRzDozKZeW")

        self.assertEqual(
            master.hwif(),
            "xpub661MyMwAqRbcFtXgS5sYJABqqG9YLmC4Q1Rdap9gSE8NqtwybGhePY2gZ29ESFjqJo"
            "Cu1Rupje8YtGqsefD265TMg7usUDFdp6W1EGMcet8")

        m0p = master.subkey(is_hardened=True)
        self.assertEqual(
            m0p.hwif(),
            "xpub68Gmy5EdvgibQVfPdqkBBCHxA5htiqg55crXYuXoQRKfDBFA1WEjWgP6LHhwBZeNK1"
            "VTsfTFUHCdrfp1bgwQ9xv5ski8PX9rL2dZXvgGDnw")
        self.assertEqual(
            m0p.hwif(as_private=True),
            "xprv9uHRZZhk6KAJC1avXpDAp4MDc3sQKNxDiPvvkX8Br5ngLNv1TxvUxt4cV1rGL5hj6K"
            "CesnDYUhd7oWgT11eZG7XnxHrnYeSvkzY7d2bhkJ7")
        self.assertEqual(master.subkey_for_path("0p").hwif(), m0p.hwif())

        pub_mp0 = master.subkey(is_hardened=True, as_private=False)
        self.assertEqual(pub_mp0.hwif(), m0p.hwif())
        self.assertEqual(master.subkey_for_path("0p.pub").hwif(), pub_mp0.hwif())

        m0p1 = m0p.subkey(i=1)
        self.assertEqual(
            m0p1.hwif(),
            "xpub6ASuArnXKPbfEwhqN6e3mwBcDTgzisQN1wXN9BJcM47sSikHjJf3UFHKkNAWbWMiGj"
            "7Wf5uMash7SyYq527Hqck2AxYysAA7xmALppuCkwQ")
        self.assertEqual(
            m0p1.hwif(as_private=True),
            "xprv9wTYmMFdV23N2TdNG573QoEsfRrWKQgWeibmLntzniatZvR9BmLnvSxqu53Kw1UmYP"
            "xLgboyZQaXwTCg8MSY3H2EU4pWcQDnRnrVA1xe8fs")
        self.assertEqual(master.subkey_for_path("0p/1").hwif(), m0p1.hwif())

        pub_m0p1 = m0p.subkey(i=1, as_private=False)
        self.assertEqual(pub_m0p1.hwif(), m0p1.hwif())
        self.assertEqual(master.subkey_for_path("0p/1.pub").hwif(), pub_m0p1.hwif())

        m0p1_1_2p = m0p1.subkey(i=2, is_hardened=True)
        self.assertEqual(
            m0p1_1_2p.hwif(),
            "xpub6D4BDPcP2GT577Vvch3R8wDkScZWzQzMMUm3PWbmWvVJrZwQY4VUNgqFJPMM3No2dF"
            "DFGTsxxpG5uJh7n7epu4trkrX7x7DogT5Uv6fcLW5")
        self.assertEqual(
            m0p1_1_2p.hwif(as_private=True),
            "xprv9z4pot5VBttmtdRTWfWQmoH1taj2axGVzFqSb8C9xaxKymcFzXBDptWmT7FwuEzG3r"
            "yjH4ktypQSAewRiNMjANTtpgP4mLTj34bhnZX7UiM")
        self.assertEqual(master.subkey_for_path("0p/1/2p").hwif(), m0p1_1_2p.hwif())

        pub_m0p1_1_2p = m0p1.subkey(i=2, as_private=False, is_hardened=True)
        self.assertEqual(pub_m0p1_1_2p.hwif(), m0p1_1_2p.hwif())
        self.assertEqual(master.subkey_for_path("0p/1/2p.pub").hwif(), pub_m0p1_1_2p.hwif())

        m0p1_1_2p_2 = m0p1_1_2p.subkey(i=2)
        self.assertEqual(
            m0p1_1_2p_2.hwif(),
            "xpub6FHa3pjLCk84BayeJxFW2SP4XRrFd1JYnxeLeU8EqN3vDfZmbqBqaGJAyiLjTAwm6Z"
            "LRQUMv1ZACTj37sR62cfN7fe5JnJ7dh8zL4fiyLHV")
        self.assertEqual(
            m0p1_1_2p_2.hwif(as_private=True),
            "xprvA2JDeKCSNNZky6uBCviVfJSKyQ1mDYahRjijr5idH2WwLsEd4Hsb2Tyh8RfQMuPh7f"
            "7RtyzTtdrbdqqsunu5Mm3wDvUAKRHSC34sJ7in334")
        self.assertEqual(master.subkey_for_path("0p/1/2p/2").hwif(), m0p1_1_2p_2.hwif())

        pub_m0p1_1_2p_2 = m0p1_1_2p.subkey(i=2, as_private=False)
        self.assertEqual(pub_m0p1_1_2p_2.hwif(), m0p1_1_2p_2.hwif())
        self.assertEqual(master.subkey_for_path("0p/1/2p/2.pub").hwif(), pub_m0p1_1_2p_2.hwif())

        m0p1_1_2p_2_1000000000 = m0p1_1_2p_2.subkey(i=1000000000)
        self.assertEqual(
            m0p1_1_2p_2_1000000000.hwif(),
            "xpub6H1LXWLaKsWFhvm6RVpEL9P4KfRZSW7abD2ttkWP3SSQvnyA8FSVqNTEcYFgJS2UaF"
            "cxupHiYkro49S8yGasTvXEYBVPamhGW6cFJodrTHy")
        self.assertEqual(
            m0p1_1_2p_2_1000000000.hwif(as_private=True),
            "xprvA41z7zogVVwxVSgdKUHDy1SKmdb533PjDz7J6N6mV6uS3ze1ai8FHa8kmHScGpWmj4"
            "WggLyQjgPie1rFSruoUihUZREPSL39UNdE3BBDu76")
        self.assertEqual(master.subkey_for_path("0p/1/2p/2/1000000000").hwif(),
                         m0p1_1_2p_2_1000000000.hwif())

        pub_m0p1_1_2p_2_1000000000 = m0p1_1_2p_2.subkey(i=1000000000, as_private=False)
        self.assertEqual(pub_m0p1_1_2p_2_1000000000.hwif(), m0p1_1_2p_2_1000000000.hwif())
        self.assertEqual(master.subkey_for_path("0p/1/2p/2/1000000000.pub").hwif(),
                         pub_m0p1_1_2p_2_1000000000.hwif())

    def test_vector_2(self):
        master = BTC.keys.bip32_seed(h2b(
            "fffcf9f6f3f0edeae7e4e1dedbd8d5d2cfccc9c6c3c0bdbab7b4b1aeaba8a5a29f9c99"
            "9693908d8a8784817e7b7875726f6c696663605d5a5754514e4b484542"))
        self.assertEqual(
            master.hwif(as_private=True),
            "xprv9s21ZrQH143K31xYSDQpPDxsXRTUcvj2iNHm5NUtrGiGG5e2DtALGdso3pGz6ssrdK"
            "4PFmM8NSpSBHNqPqm55Qn3LqFtT2emdEXVYsCzC2U")

        self.assertEqual(
            master.hwif(),
            "xpub661MyMwAqRbcFW31YEwpkMuc5THy2PSt5bDMsktWQcFF8syAmRUapSCGu8ED9W6oDM"
            "Sgv6Zz8idoc4a6mr8BDzTJY47LJhkJ8UB7WEGuduB")

        m0 = master.subkey()
        self.assertEqual(
            m0.hwif(),
            "xpub69H7F5d8KSRgmmdJg2KhpAK8SR3DjMwAdkxj3ZuxV27CprR9LgpeyGmXUbC6wb7ERf"
            "vrnKZjXoUmmDznezpbZb7ap6r1D3tgFxHmwMkQTPH")
        self.assertEqual(
            m0.hwif(as_private=True),
            "xprv9vHkqa6EV4sPZHYqZznhT2NPtPCjKuDKGY38FBWLvgaDx45zo9WQRUT3dKYnjwih2y"
            "JD9mkrocEZXo1ex8G81dwSM1fwqWpWkeS3v86pgKt")
        pub_m0 = master.subkey(as_private=False)
        self.assertEqual(pub_m0.hwif(), m0.hwif())

        m0_2147483647p = m0.subkey(i=2147483647, is_hardened=True)
        self.assertEqual(
            m0_2147483647p.hwif(),
            "xpub6ASAVgeehLbnwdqV6UKMHVzgqAG8Gr6riv3Fxxpj8ksbH9ebxaEyBLZ85ySDhKiLDB"
            "rQSARLq1uNRts8RuJiHjaDMBU4Zn9h8LZNnBC5y4a")
        self.assertEqual(
            m0_2147483647p.hwif(as_private=True),
            "xprv9wSp6B7kry3Vj9m1zSnLvN3xH8RdsPP1Mh7fAaR7aRLcQMKTR2vidYEeEg2mUCTAwC"
            "d6vnxVrcjfy2kRgVsFawNzmjuHc2YmYRmagcEPdU9")
        pub_m0_2147483647p = m0.subkey(i=2147483647, is_hardened=True, as_private=False)
        self.assertEqual(pub_m0_2147483647p.hwif(), m0_2147483647p.hwif())

        m0_2147483647p_1 = m0_2147483647p.subkey(i=1)
        self.assertEqual(
            m0_2147483647p_1.hwif(),
            "xpub6DF8uhdarytz3FWdA8TvFSvvAh8dP3283MY7p2V4SeE2wyWmG5mg5EwVvmdMVCQcoN"
            "JxGoWaU9DCWh89LojfZ537wTfunKau47EL2dhHKon")
        self.assertEqual(
            m0_2147483647p_1.hwif(as_private=True),
            "xprv9zFnWC6h2cLgpmSA46vutJzBcfJ8yaJGg8cX1e5StJh45BBciYTRXSd25UEPVuesF9"
            "yog62tGAQtHjXajPPdbRCHuWS6T8XA2ECKADdw4Ef")
        pub_m0_2147483647p_1 = m0_2147483647p.subkey(i=1, as_private=False)
        self.assertEqual(pub_m0_2147483647p_1.hwif(), m0_2147483647p_1.hwif())
        pub_m0_2147483647p_1 = pub_m0_2147483647p.subkey(i=1, as_private=False)
        self.assertEqual(pub_m0_2147483647p_1.hwif(), m0_2147483647p_1.hwif())

        m0_2147483647p_1_2147483646p = m0_2147483647p_1.subkey(i=2147483646, is_hardened=True)
        self.assertEqual(
            m0_2147483647p_1_2147483646p.hwif(),
            "xpub6ERApfZwUNrhLCkDtcHTcxd75RbzS1ed54G1LkBUHQVHQKqhMkhgbmJbZRkrgZw4ko"
            "xb5JaHWkY4ALHY2grBGRjaDMzQLcgJvLJuZZvRcEL")
        self.assertEqual(
            m0_2147483647p_1_2147483646p.hwif(as_private=True),
            "xprvA1RpRA33e1JQ7ifknakTFpgNXPmW2YvmhqLQYMmrj4xJXXWYpDPS3xz7iAxn8L39nj"
            "GVyuoseXzU6rcxFLJ8HFsTjSyQbLYnMpCqE2VbFWc")
        pub_m0_2147483647p_1_2147483646p = m0_2147483647p_1.subkey(i=2147483646, as_private=False, is_hardened=True)
        self.assertEqual(pub_m0_2147483647p_1_2147483646p.hwif(), m0_2147483647p_1_2147483646p.hwif())

        m0_2147483647p_1_2147483646p_2 = m0_2147483647p_1_2147483646p.subkey(i=2)
        self.assertEqual(m0_2147483647p_1_2147483646p_2.wif(), "L3WAYNAZPxx1fr7KCz7GN9nD5qMBnNiqEJNJMU1z9MMaannAt4aK")
        self.assertEqual(
            m0_2147483647p_1_2147483646p_2.hwif(),
            "xpub6FnCn6nSzZAw5Tw7cgR9bi15UV96gLZhjDstkXXxvCLsUXBGXPdSnLFbdpq8p9HmGs"
            "ApME5hQTZ3emM2rnY5agb9rXpVGyy3bdW6EEgAtqt")
        self.assertEqual(
            m0_2147483647p_1_2147483646p_2.hwif(as_private=True),
            "xprvA2nrNbFZABcdryreWet9Ea4LvTJcGsqrMzxHx98MMrotbir7yrKCEXw7nadnHM8Dq3"
            "8EGfSh6dqA9QWTyefMLEcBYJUuekgW4BYPJcr9E7j")
        pub_m0_2147483647p_1_2147483646p_2 = m0_2147483647p_1_2147483646p.subkey(i=2, as_private=False)
        self.assertEqual(pub_m0_2147483647p_1_2147483646p_2.hwif(), m0_2147483647p_1_2147483646p_2.hwif())
        pub_m0_2147483647p_1_2147483646p_2 = pub_m0_2147483647p_1_2147483646p.subkey(i=2, as_private=False)
        self.assertEqual(pub_m0_2147483647p_1_2147483646p_2.hwif(), m0_2147483647p_1_2147483646p_2.hwif())
        self.assertEqual(master.subkey_for_path("0/2147483647p/1/2147483646p/2").hwif(),
                         m0_2147483647p_1_2147483646p_2.hwif())
        self.assertEqual(master.subkey_for_path("0/2147483647p/1/2147483646p/2.pub").hwif(),
                         pub_m0_2147483647p_1_2147483646p_2.hwif())

    def test_testnet(self):
        # WARNING: these values have not been verified independently. TODO: do so
        master = XTN.keys.bip32_seed(h2b("000102030405060708090a0b0c0d0e0f"))
        self.assertEqual(
            master.hwif(as_private=True),
            "tprv8ZgxMBicQKsPeDgjzdC36fs6bMjGApWDNLR9erAXMs5skhMv36j9MV5ecvfavji5kh"
            "qjWaWSFhN3YcCUUdiKH6isR4Pwy3U5y5egddBr16m")
        self.assertEqual(master.address(), "mkHGce7dctSxHgaWSSbmmrRWsZfzz7MxMk")
        self.assertEqual(master.wif(), "cVPXTF2TnozE1PenpP3x9huctiATZmp27T9Ue1d8nqLSExoPwfN5")

    def test_streams(self):
        m0 = BTC.keys.bip32_seed(b"foo bar baz")
        pm0 = m0.public_copy()
        self.assertEqual(m0.hwif(), pm0.hwif())
        m1 = m0.subkey()
        pm1 = pm0.subkey()
        for i in range(4):
            m = m1.subkey(i=i)
            pm = pm1.subkey(i=i)
            self.assertEqual(m.hwif(), pm.hwif())
            self.assertEqual(m.address(), pm.address())
            m2 = BTC.parse.secret(m.hwif(as_private=True))
            m3 = m2.public_copy()
            self.assertEqual(m.hwif(as_private=True), m2.hwif(as_private=True))
            self.assertEqual(m.hwif(), m3.hwif())
            print(m.hwif(as_private=True))
            for j in range(2):
                k = m.subkey(i=j)
                k2 = BTC.parse.secret(k.hwif(as_private=True))
                k3 = BTC.parse.secret(k.hwif())
                k4 = k.public_copy()
                self.assertEqual(k.hwif(as_private=True), k2.hwif(as_private=True))
                self.assertEqual(k.hwif(), k2.hwif())
                self.assertEqual(k.hwif(), k3.hwif())
                self.assertEqual(k.hwif(), k4.hwif())
                print("   %s %s" % (k.address(), k.wif()))

    def test_public_subkey(self):
        my_prv = BTC.keys.bip32_seed(b"foo")
        uag = my_prv.subkey(i=0, is_hardened=True, as_private=True)
        self.assertEqual(None, uag.subkey(i=0, as_private=False).secret_exponent())

        with self.assertRaises(ValueError) as cm:
            my_prv.subkey(i=-1)

        err = cm.exception
        self.assertEqual(err.args, ("i can't be negative", ))

        for p in ('-1', '0/-1', '0H/-1'):
            with self.assertRaises(ValueError) as cm:
                my_prv.subkey_for_path(p)

            err = cm.exception
            self.assertEqual(err.args, ("i can't be negative", ))

        self.assertRaises(ValueError, list, my_prv.subkeys('-1'))
        self.assertRaises(ValueError, list, my_prv.subkeys('-1-0'))

    def test_repr(self):
        key = XTN.keys.private(secret_exponent=273)
        wallet = XTN.keys.bip32_seed(bytes(key.wif().encode('utf8')))

        address = wallet.address()
        pub_k = XTN.parse.address(address)
        self.assertEqual(repr(pub_k),  '<myb5gZNXePNf2E2ksrjnHRFCwyuvt7oEay>')

        wif = wallet.wif()
        priv_k = XTN.parse.secret(wif)
        self.assertEqual(repr(priv_k),
                         'private_for <XTNSEC:03ad094b1dc9fdce5d3648ca359b4e210a89d049532fdd39d9ccdd8ca393ac82f4>')


if __name__ == '__main__':
    unittest.main()




from .SolutionChecker import BgoldSolutionChecker
from .Solver import BgoldSolver

from pycoin.coins.bitcoin.Tx import Tx as BaseTx


class Tx(BaseTx):
    Solver = BgoldSolver
    SolutionChecker = BgoldSolutionChecker



from pycoin.networks.bitcoinish import create_bitcoinish_network

network = create_bitcoinish_network(
    network_name="Chaucha", symbol="CHA", subnet_name="mainnet",
    wif_prefix_hex="d8", address_prefix_hex="58", pay_to_script_prefix_hex="50",
    bip32_prv_prefix_hex="0488ADE4", bip32_pub_prefix_hex="0488B21E",
    magic_header_hex="AAA226A9", default_port=21663,
    dns_bootstrap=[
        "condor420.chaucha.cl", "huemul69.chaucha.cl",
    ])



from .satoshi_streamer import SATOSHI_STREAMER

parse_struct = SATOSHI_STREAMER.parse_struct
parse_as_dict = SATOSHI_STREAMER.parse_as_dict
stream_struct = SATOSHI_STREAMER.stream_struct
pack_struct = SATOSHI_STREAMER.pack_struct



from pycoin.networks.bitcoinish import create_bitcoinish_network

network = create_bitcoinish_network(
    symbol="BTDX", network_name="Bitcloud", subnet_name="mainnet",
    wif_prefix_hex="99", sec_prefix="BTDXSEC:", address_prefix_hex="19", pay_to_script_prefix_hex="05",
    bip32_prv_prefix_hex="0488ADE4", bip32_pub_prefix_hex="0488B21E",
    magic_header_hex="E4E8BDFD", default_port=8329,
    dns_bootstrap=[
        "seed.bitcloud.network"
    ])




from pycoin.coins.SolutionChecker import ScriptError
from pycoin.satoshi import intops, stackops, checksigops, miscops
from pycoin.satoshi import errno

from .ScriptStreamer import BitcoinScriptStreamer


def _make_bad_instruction(v):
    def f(vm_state):
        raise ScriptError("invalid instruction x%02x at %d" % (v, vm_state.pc), errno.BAD_OPCODE)
    return f


def _collect_opcodes(module):
    d = {}
    for k in dir(module):
        if k.startswith("do_OP"):
            d[k[3:]] = getattr(module, k)
    return d


def _no_op(vm):
    pass


def make_instruction_lookup(opcode_pairs):
    OPCODE_DATA_LIST = list(BitcoinScriptStreamer.data_opcodes)

    # start with all opcodes invalid
    instruction_lookup = [_make_bad_instruction(i) for i in range(256)]

    for i in OPCODE_DATA_LIST:
        instruction_lookup[i] = _no_op
    opcode_lookups = {}
    opcode_lookups.update(_collect_opcodes(checksigops))
    opcode_lookups.update(_collect_opcodes(intops))
    opcode_lookups.update(_collect_opcodes(stackops))
    opcode_lookups.update(_collect_opcodes(miscops))
    opcode_lookups.update(miscops.extra_opcodes())
    for opcode_name, opcode_value in opcode_pairs:
        if opcode_name in opcode_lookups:
            instruction_lookup[opcode_value] = opcode_lookups[opcode_name]
    return instruction_lookup



from pycoin.coins.groestlcoin.hash import groestlHash
from pycoin.coins.groestlcoin.parse import GRSParseAPI
from pycoin.coins.groestlcoin.Block import Block as GrsBlock
from pycoin.coins.groestlcoin.Tx import Tx as GrsTx
from pycoin.encoding.b58 import b2a_base58
from pycoin.encoding.hexbytes import h2b
from pycoin.networks.bitcoinish import create_bitcoinish_network


network = create_bitcoinish_network(
    symbol="GRS", network_name="Groestlcoin", subnet_name="mainnet", tx=GrsTx, block=GrsBlock,
    wif_prefix_hex="80", sec_prefix="GRSSEC:", address_prefix_hex="24", pay_to_script_prefix_hex="05",
    bip32_prv_prefix_hex="0488ade4", bip32_pub_prefix_hex="0488B21E", bech32_hrp="grs",
    magic_header_hex="F9BEB4D9", default_port=1331,
    parse_api_class=GRSParseAPI,
    dns_bootstrap=[
        "groestlcoin.org", "electrum1.groestlcoin.org", "electrum2.groestlcoin.org",
        "jswallet.groestlcoin.org", "groestlsight.groestlcoin.org"
    ]
)

# monkey patches
_wif_prefix = h2b("80")
_bip32_prv_prefix = h2b("0488ade4")
_bip32_pub_prefix = h2b("0488B21E")


def b2a_hashed_base58_grs(data):
    return b2a_base58(data + groestlHash(data)[:4])


def bip32_as_string(blob, as_private):
    prefix = _bip32_prv_prefix if as_private else _bip32_pub_prefix
    return b2a_hashed_base58_grs(prefix + blob)


def wif_for_blob(blob):
    return b2a_hashed_base58_grs(_wif_prefix + blob)


network.address.b2a = b2a_hashed_base58_grs
network.bip32_as_string = bip32_as_string
network.wif_for_blob = wif_for_blob

# Cause parsing to fail and tests to skip.
try:
    import groestlcoin_hash  # noqa
except ImportError:
    network.Key = None

    def none_parser(*args, **kwargs):
        return None

    for attr in "hierarchical_key private_key public_key address".split():
        setattr(network.parse, attr, none_parser)



from pycoin.networks.bitcoinish import create_bitcoinish_network

network = create_bitcoinish_network(
    symbol="BC", network_name="Blackcoin", subnet_name="mainnet",
    wif_prefix_hex="99", address_prefix_hex="19",
    bip32_prv_prefix_hex="02cfbf60", bip32_pub_prefix_hex="02cfbede")



from ...intbytes import byte2int, indexbytes

from ..SolutionChecker import SolutionChecker

from pycoin.satoshi.flags import VERIFY_P2SH

from .ScriptTools import BitcoinScriptTools


OP_EQUAL = BitcoinScriptTools.int_for_opcode("OP_EQUAL")
OP_HASH160 = BitcoinScriptTools.int_for_opcode("OP_HASH160")


class P2SChecker(SolutionChecker):

    @classmethod
    def is_pay_to_script_hash(class_, script_public_key):
        return (len(script_public_key) == 23 and byte2int(script_public_key) == OP_HASH160 and
                indexbytes(script_public_key, -1) == OP_EQUAL)

    @classmethod
    def script_hash_from_script(class_, puzzle_script):
        if class_.is_pay_to_script_hash(puzzle_script):
            return puzzle_script[2:-1]
        return False

    def p2s_program_tuple(self, tx_context, puzzle_script, solution_stack, flags, sighash_f):
        if flags & VERIFY_P2SH and self.is_pay_to_script_hash(puzzle_script):
            self._check_script_push_only(tx_context.solution_script)
            puzzle_script, solution_stack = solution_stack[-1], solution_stack[:-1]
            return puzzle_script, solution_stack, flags & ~VERIFY_P2SH, sighash_f




from pycoin.intbytes import iterbytes, byte2int


def _to_bytes(v, length, byteorder="big"):
    """This is the same functionality as ``int.to_bytes`` in python 3"""
    return v.to_bytes(length, byteorder=byteorder)


def _from_bytes(bytes, byteorder="big", signed=False):
    """This is the same functionality as ``int.from_bytes`` in python 3"""
    return int.from_bytes(bytes, byteorder=byteorder, signed=signed)


if hasattr(int, "to_bytes"):
    to_bytes = _to_bytes
    from_bytes = _from_bytes
else:
    def to_bytes(v, length, byteorder="big"):
        """This is the same functionality as ``int.to_bytes`` in python 3"""
        ba = bytearray()
        for i in range(length):
            mod = v & 0xff
            v >>= 8
            ba.append(mod)
        if byteorder == "big":
            ba.reverse()
        return bytes(ba)

    def from_bytes(bytes, byteorder="big", signed=False):
        """This is the same functionality as ``int.from_bytes`` in python 3"""
        if byteorder != "big":
            bytes = reversed(bytes)
        v = 0
        for c in iterbytes(bytes):
            v <<= 8
            v += c
        if signed and byte2int(bytes) & 0x80:
            v = v - (1 << (8*len(bytes)))
        return v



from pycoin.networks.bitcoinish import create_bitcoinish_network


network = create_bitcoinish_network(
    symbol="XTN", network_name="Bitcoin", subnet_name="testnet3",
    wif_prefix_hex="ef", sec_prefix="XTNSEC:", address_prefix_hex="6f", pay_to_script_prefix_hex="c4",
    bip32_prv_prefix_hex="04358394", bip32_pub_prefix_hex="043587CF", bech32_hrp="tb",
    magic_header_hex="0B110907", default_port=18333,
    dns_bootstrap=[
        "bitcoin.petertodd.org", "testnet-seed.bitcoin.petertodd.org",
        "bluematt.me", "testnet-seed.bluematt.me"
    ])




import io


class Streamer(object):
    def __init__(self):
        self.parse_lookup = {}
        self.stream_lookup = {}

    def register_functions(self, lookup):
        for c, v in lookup:
            parse_f, stream_f = v
            self.parse_lookup[c] = parse_f
            self.stream_lookup[c] = stream_f

    def register_array_count_parse(self, array_count_parse_f):
        self.array_count_parse_f = array_count_parse_f

    def parse_struct(self, fmt, f):
        items = []
        i = 0
        while i < len(fmt):
            c = fmt[i]
            if c == "[":
                end = fmt.find("]", i)
                if end < 0:
                    raise ValueError("no closing ] character")
                subfmt = fmt[i+1:end]
                count = self.array_count_parse_f(f)
                array = []
                for j in range(count):
                    if len(subfmt) == 1:
                        array.append(self.parse_struct(subfmt, f)[0])
                    else:
                        array.append(self.parse_struct(subfmt, f))
                items.append(tuple(array))
                i = end
            else:
                items.append(self.parse_lookup[c](f))
            i += 1
        return tuple(items)

    def parse_as_dict(self, attribute_list, pack_list, f):
        return dict(list(zip(attribute_list, self.parse_struct(pack_list, f))))

    def stream_struct(self, fmt, f, *args):
        for c, v in zip(fmt, args):
            self.stream_lookup[c](f, v)

    def unpack_struct(self, fmt, b):
        return self.parse_struct(fmt, io.BytesIO(b))

    def pack_struct(self, fmt, *args):
        b = io.BytesIO()
        self.stream_struct(fmt, b, *args)
        return b.getvalue()



#!/usr/bin/env python

import sys

from pycoin.encoding.hexbytes import h2b
from pycoin.symbols.btc import network


def main():
    if len(sys.argv) != 4:
        print("usage: %s tx-hex-file-path wif-file-path p2sh-file-path" % sys.argv[0])
        sys.exit(-1)

    # get the tx
    with open(sys.argv[1], "r") as f:
        tx_hex = f.readline().strip()
    tx = network.tx.from_hex(tx_hex)

    # get the WIF
    with open(sys.argv[2], "r") as f:
        wif = f.readline().strip()
    assert network.parse.wif(wif) is not None

    # create the p2sh_lookup
    with open(sys.argv[3], "r") as f:
        p2sh_script_hex = f.readline().strip()
    p2sh_script = h2b(p2sh_script_hex)

    # build a dictionary of script hashes to scripts
    p2sh_lookup = network.tx.solve.build_p2sh_lookup([p2sh_script])

    # sign the transaction with the given WIF
    network.tx_utils.sign_tx(tx, wifs=[wif], p2sh_lookup=p2sh_lookup)

    bad_solution_count = tx.bad_solution_count()
    print("tx %s now has %d bad solution(s)" % (tx.id(), bad_solution_count))

    include_unspents = (bad_solution_count > 0)
    print("Here is the tx as hex:\n%s" % tx.as_hex(include_unspents=include_unspents))


if __name__ == '__main__':
    main()



import logging
import platform
import struct
import os

from pycoin.block import Block
from pycoin.blockchain.BlockChain import BlockChain
from pycoin.encoding.hexbytes import h2b


logger = logging.getLogger(__file__)


# f9beb4d9 (big-endian magic)
# 011d (little-endian size)
# (block data)
# f9beb4d9 (big-endian magic)
# (etc)


class Blockfiles(object):
    def __init__(self, base_dir=None, start_info=(0, 0), MAGIC=h2b("f9beb4d9")):
        if base_dir is None:
            base_dir = self.default_base()
        self.base_dir = base_dir
        self._file_index = None
        self._magic = MAGIC
        self.jump_to(start_info)

    def jump_to(self, start_info):
        file_index, offset = start_info
        if self._file_index != file_index:
            self._file_index = file_index
            full_path = self._path_for_file_index()
            self.f = open(full_path, "rb")
        self.f.seek(offset)

    def close(self):
        self.f.close()

    def default_base(self):
        LOOKUP = dict(Darwin="~/Library/Application Support/Bitcoin/", Linux="~/.bitcoin/")
        system = platform.system()
        path = LOOKUP.get(system)
        if path is None:
            raise ValueError("unknown base path for system %s; you should submit a patch!" % system)
        return os.path.expanduser(path)

    def read(self, N):
        d = self.f.read(N)
        if len(d) >= N:
            return d
        if self._next_file():
            d1 = d + self.f.read(N-len(d))
            return d1
        return b""

    def skip(self, offset):
        while 1:
            cur_pos = self.f.tell()
            self.f.seek(offset, 1)
            new_pos = self.f.tell()
            offset += cur_pos - new_pos
            if offset <= 0:
                break
            if not self.next_file():
                break

    def _path_for_file_index(self):
        return os.path.join(self.base_dir, "blocks", "blk%05d.dat" % self._file_index)

    def _next_file(self):
        self._file_index += 1
        full_path = self._path_for_file_index()
        if not os.path.exists(full_path):
            return False
        self.f.close()
        self.f = open(full_path, "r+b")
        return True

    def next_offset(self, current_offset):
        self.jump_to(current_offset)
        magic = self.read(4)
        if magic == b"\0\0\0\0":
            if not self._next_file():
                return None
            magic = self.read(4)
        if len(magic) == 0:
            return None
        if magic != self._magic:
            offset_info = self.offset_info()
            logger.error("bad magic: %s at %s", magic, offset_info)
            raise ValueError("bad magic at %s" % str(offset_info))
        size = struct.unpack("<L", self.read(4))[0]
        block_offset = self.offset_info()
        self.skip(size)
        next_offset = self.offset_info()
        return block_offset, next_offset

    def offset_info(self):
        return self._file_index, self.f.tell()


def locked_blocks_iterator(blockfile, start_info=(0, 0), cached_headers=50, batch_size=50):
    """
    This method loads blocks from disk, skipping any orphan blocks.
    """
    f = blockfile
    current_state = []

    def change_state(bc, ops):
        for op, bh, work in ops:
            if op == 'add':
                current_state.append(bh)
                pass
            else:
                current_state.pop()
    bc = BlockChain()
    bc.add_change_callback(change_state)
    bhs = []
    index = 0
    info_offset = start_info
    while 1:
        v = blockfile.next_offset(info_offset)
        if v is None:
            break
        block_offset, info_offset = v
        f.jump_to(block_offset)
        bh = Block.parse_as_header(f)
        bh.info = block_offset

        bhs.append(bh)
        if len(bhs) > batch_size:
            bc.add_headers(bhs)
            bhs = []
            if len(current_state) > cached_headers:
                for bh in current_state[:cached_headers]:
                    bh.index = index
                    yield bh
                    index += 1
                    bc.lock_to_index(index)
                current_state = current_state[cached_headers:]




import threading
import unittest

from pycoin.encoding.hexbytes import h2b_rev
from pycoin.services import providers
from pycoin.services.blockchain_info import BlockchainInfoProvider
from pycoin.services.blockcypher import BlockcypherProvider
from pycoin.services.blockexplorer import BlockExplorerProvider
from pycoin.services.chain_so import ChainSoProvider
from pycoin.services.insight import InsightProvider


BLOCK_0_HASH = h2b_rev("000000000019d6689c085ae165831e934ff763ae46a2a6c172b3f1b60a8ce26f")
BLOCK_1_HASH = h2b_rev("00000000839a8e6886ab5951d76f411475428afc90947ee320161bbf18eb6048")


tx_id_for_net = {
    "BTC": ["b958e4a3ccd5bc8fe0ff6fafd635199313e347b88a8102040c05dd123f32a4f3",
            "d1ef46055a84fd02ee82580d691064780def18614d98646371c3448ca20019ac",
            "69916297f7adde13457b8244e2d704966097e9519ec8fd6f2e7af8c2a60f70f2",
            "c586389e5e4b3acb9d6c8be1c19ae8ab2795397633176f5a6442a261bbdefc3a"],
    "XTN": ["4586e67ee5adcdbc97ed3d2a026ee8703df2ed3553854c186c216e90cd761b69"],
    "DOGE": ["ed7df4e7506ac8447b6983c8ad79da1af86cddda0ff012f7db83e664f61ef6cf"],
    "XDT": ["19dd5c3423e606b5b5dd30b070688bdf9af27fa736e8f3aeb2b68d92a50e67ef"],
}


class ServicesTest(unittest.TestCase):
    def test_env(self):
        CS = "blockchain.info blockexplorer.com chain.so insight:https://hostname/url"
        provider_list = providers.providers_for_config_string(CS, "BTC")
        self.assertEqual(len(provider_list), len(CS.split()))

    def test_thread_provider(self):
        p_list_1 = providers.providers_for_config_string("blockchain.info", "BTC")
        p_list_2 = providers.providers_for_config_string("blockexplorer.com", "BTC")
        providers.set_default_providers_for_netcode("BTC", p_list_1)
        self.assertEqual(providers.get_default_providers_for_netcode("BTC"), p_list_1)
        # use a dictionary so it can be mutable in the subthread
        d = {"is_ok": False}

        def subthread():
            providers.set_default_providers_for_netcode("BTC", [])
            self.assertEqual(providers.get_default_providers_for_netcode("BTC"), [])
            providers.set_default_providers_for_netcode("BTC", p_list_2)
            self.assertEqual(providers.get_default_providers_for_netcode("BTC"), p_list_2)
            d["is_ok"] = True
        t = threading.Thread(target=subthread)
        t.start()
        t.join()
        self.assertTrue(d["is_ok"])
        self.assertEqual(providers.get_default_providers_for_netcode("BTC"), p_list_1)

    def check_provider_tx_for_tx_hash(self, p, networks):
        for net in networks:
            b = p(net)
            for tx_id in tx_id_for_net[net]:
                tx = b.tx_for_tx_hash(h2b_rev(tx_id))
                self.assertEqual(tx.id(), tx_id)

    def test_BitcoindProvider(self):
        # not sure what to do here, as there is no open bitcoind provider I know of
        pass

    def test_BlockchainInfo(self):
        self.check_provider_tx_for_tx_hash(BlockchainInfoProvider, ["BTC"])

    @unittest.skip("this test is not working for some reason")
    def test_BlockCypherProvider(self):
        self.check_provider_tx_for_tx_hash(BlockcypherProvider, ["BTC", "XTN"])

    @unittest.skip("this test is not working for some reason")
    def test_BlockExplorerProvider(self):
        self.check_provider_tx_for_tx_hash(BlockExplorerProvider, ["BTC"])

    @unittest.skip("this test is causing problems in travis-ci because chain_so thinks it's a DOS attack")
    def test_ChainSoProvider(self):
        self.check_provider_tx_for_tx_hash(ChainSoProvider, ["BTC", "XTN", "DOGE", "XDT"])

    def test_InsightProvider(self):
        provider = InsightProvider("http://insight.bitpay.com/api/")
        self.check_provider_tx_for_tx_hash(lambda x: provider, ["BTC"])
        provider.get_blockchain_tip()
        h = provider.get_blockheader(BLOCK_1_HASH)
        assert h.previous_block_hash == BLOCK_0_HASH
        height = provider.get_block_height(BLOCK_1_HASH)
        assert height == 1


def main():
    unittest.main()


if __name__ == "__main__":
    main()



from .ScriptTools import BitcoinScriptTools
from .VM import BitcoinVM

from ...encoding.bytes32 import from_bytes_32

from pycoin.satoshi import errno
from pycoin.satoshi.flags import (
    SIGHASH_NONE, SIGHASH_SINGLE, SIGHASH_ANYONECANPAY,
    VERIFY_P2SH, VERIFY_SIGPUSHONLY, VERIFY_CLEANSTACK,
    VERIFY_WITNESS, VERIFY_MINIMALIF, VERIFY_WITNESS_PUBKEYTYPE
)

from .SegwitChecker import SegwitChecker
from .P2SChecker import P2SChecker


class TxContext(object):
    pass


class BitcoinSolutionChecker(SegwitChecker, P2SChecker):
    VM = BitcoinVM
    ScriptTools = BitcoinScriptTools

    DEFAULT_FLAGS = VERIFY_P2SH | VERIFY_WITNESS

    def __init__(self, tx):
        self.tx = tx
        # self.sighash_cache = {}

    def _delete_signature(self, script, sig_blob):
        """
        Returns a script with the given subscript removed. The subscript
        must appear in the main script aligned to opcode boundaries for it
        to be removed.
        """
        subscript = self.ScriptTools.compile_push_data_list([sig_blob])
        new_script = bytearray()
        pc = 0
        for opcode, data, pc, new_pc in self.ScriptTools.get_opcodes(script):
            section = script[pc:new_pc]
            if section != subscript:
                new_script.extend(section)
        return bytes(new_script)

    def _make_sighash_f(self, tx_in_idx):

        def sig_for_hash_type_f(hash_type, sig_blobs, vm):
            script = vm.script[vm.begin_code_hash:]
            for sig_blob in sig_blobs:
                script = self._delete_signature(script, sig_blob)
            return self._signature_hash(script, tx_in_idx, hash_type)

        return sig_for_hash_type_f

    def _solution_script_to_stack(self, tx_context, flags, traceback_f):
        if flags & VERIFY_SIGPUSHONLY:
            self._check_script_push_only(tx_context.solution_script)

        # never use VERIFY_MINIMALIF or VERIFY_WITNESS_PUBKEYTYPE except in segwit
        f1 = flags & ~(VERIFY_MINIMALIF | VERIFY_WITNESS_PUBKEYTYPE)

        vm = self.VM(tx_context.solution_script, tx_context, self._make_sighash_f(tx_context.tx_in_idx), f1)

        vm.is_solution_script = True
        vm.traceback_f = traceback_f

        solution_stack = vm.eval_script()
        return solution_stack

    def _check_script_push_only(self, script):
        scriptStreamer = self.VM.ScriptStreamer
        pc = 0
        while pc < len(script):
            opcode, data, pc, is_ok = scriptStreamer.get_opcode(script, pc)
            if opcode not in scriptStreamer.data_opcodes:
                raise self.ScriptError("signature has non-push opcodes", errno.SIG_PUSHONLY)

    def _tx_in_for_idx(self, idx, tx_in, tx_out_script, unsigned_txs_out_idx):
        if idx == unsigned_txs_out_idx:
            return self.tx.TxIn(tx_in.previous_hash, tx_in.previous_index, tx_out_script, tx_in.sequence)
        return self.tx.TxIn(tx_in.previous_hash, tx_in.previous_index, b'', tx_in.sequence)

    @classmethod
    def delete_subscript(class_, script, subscript):
        """
        Returns a script with the given subscript removed. The subscript
        must appear in the main script aligned to opcode boundaries for it
        to be removed.
        """
        new_script = bytearray()
        pc = 0
        for opcode, data, pc, new_pc in class_.ScriptTools.get_opcodes(script):
            section = script[pc:new_pc]
            if section != subscript:
                new_script.extend(section)
        return bytes(new_script)

    def _signature_hash(self, tx_out_script, unsigned_txs_out_idx, hash_type):
        """
        Return the canonical hash for a transaction. We need to
        remove references to the signature, since it's a signature
        of the hash before the signature is applied.

        :param tx_out_script: the script the coins for unsigned_txs_out_idx are coming from
        :param unsigned_txs_out_idx: where to put the tx_out_script
        :param hash_type: one of SIGHASH_NONE, SIGHASH_SINGLE, SIGHASH_ALL,
            optionally bitwise or'ed with SIGHASH_ANYONECANPAY
        """

        # In case concatenating two scripts ends up with two codeseparators,
        # or an extra one at the end, this prevents all those possible incompatibilities.
        tx_out_script = self.delete_subscript(tx_out_script, self.ScriptTools.compile("OP_CODESEPARATOR"))

        # blank out other inputs' signatures
        txs_in = [self._tx_in_for_idx(i, tx_in, tx_out_script, unsigned_txs_out_idx)
                  for i, tx_in in enumerate(self.tx.txs_in)]
        txs_out = self.tx.txs_out

        # Blank out some of the outputs
        if (hash_type & 0x1f) == SIGHASH_NONE:
            # Wildcard payee
            txs_out = []

            # Let the others update at will
            for i in range(len(txs_in)):
                if i != unsigned_txs_out_idx:
                    txs_in[i].sequence = 0

        elif (hash_type & 0x1f) == SIGHASH_SINGLE:
            # This preserves the ability to validate existing legacy
            # transactions which followed a buggy path in Satoshi's
            # original code.
            if unsigned_txs_out_idx >= len(txs_out):
                # This should probably be moved to a constant, but the
                # likelihood of ever getting here is already really small
                # and getting smaller
                return (1 << 248)

            # Only lock in the txout payee at same index as txin; delete
            # any outputs after this one and set all outputs before this
            # one to "null" (where "null" means an empty script and a
            # value of -1)
            txs_out = [self.tx.TxOut(0xffffffffffffffff, b'')] * unsigned_txs_out_idx
            txs_out.append(self.tx.txs_out[unsigned_txs_out_idx])

            # Let the others update at will
            for i in range(len(txs_in)):
                if i != unsigned_txs_out_idx:
                    txs_in[i].sequence = 0

        # Blank out other inputs completely, not recommended for open transactions
        if hash_type & SIGHASH_ANYONECANPAY:
            txs_in = [txs_in[unsigned_txs_out_idx]]

        tmp_tx = self.tx.__class__(self.tx.version, txs_in, txs_out, self.tx.lock_time)
        return from_bytes_32(tmp_tx.hash(hash_type=hash_type))

    def tx_context_for_idx(self, tx_in_idx):
        """
        solution_script: alleged solution to the puzzle_script
        puzzle_script: the script protecting the coins
        """
        tx_in = self.tx.txs_in[tx_in_idx]

        tx_context = TxContext()
        tx_context.lock_time = self.tx.lock_time
        tx_context.version = self.tx.version
        tx_context.puzzle_script = b'' if self.tx.missing_unspent(tx_in_idx) else self.tx.unspents[tx_in_idx].script
        tx_context.solution_script = tx_in.script
        tx_context.witness_solution_stack = tx_in.witness
        tx_context.sequence = tx_in.sequence
        tx_context.tx_in_idx = tx_in_idx
        return tx_context

    def check_solution(self, tx_context, flags=None, traceback_f=None):
        """
        tx_context: information about the transaction that the VM may need
        flags: gives the VM hints about which additional constraints to check
        """

        for t in self.puzzle_and_solution_iterator(tx_context, flags=flags, traceback_f=traceback_f):
            puzzle_script, solution_stack, flags, sighash_f = t

            vm = self.VM(puzzle_script, tx_context, sighash_f, flags=flags, initial_stack=solution_stack[:])

            vm.is_solution_script = False
            vm.traceback_f = traceback_f

            stack = vm.eval_script()
            if len(stack) == 0 or not vm.bool_from_script_bytes(stack[-1]):
                raise self.ScriptError("eval false", errno.EVAL_FALSE)

        if flags & VERIFY_CLEANSTACK and len(stack) != 1:
            raise self.ScriptError("stack not clean after evaluation", errno.CLEANSTACK)

    def puzzle_and_solution_iterator(self, tx_context, flags=None, traceback_f=None):
        if flags is None:
            flags = self.DEFAULT_FLAGS

        solution_stack = self._solution_script_to_stack(tx_context, flags=flags, traceback_f=traceback_f)
        puzzle_script = tx_context.puzzle_script

        flags_1 = flags & ~(VERIFY_MINIMALIF | VERIFY_WITNESS_PUBKEYTYPE)

        sighash_f = self._make_sighash_f(tx_context.tx_in_idx)
        yield puzzle_script, solution_stack, flags_1, sighash_f

        p2sh_tuple = self.p2s_program_tuple(tx_context, puzzle_script, solution_stack, flags_1, sighash_f)
        if p2sh_tuple:
            yield p2sh_tuple
            puzzle_script, solution_stack = p2sh_tuple[:2]

        is_p2sh = p2sh_tuple is not None
        witness_tuple = self.witness_program_tuple(tx_context, puzzle_script, solution_stack, flags, is_p2sh)
        if witness_tuple:
            yield witness_tuple



from __future__ import print_function

import datetime

from pycoin.convention import satoshi_to_mbtc
from pycoin.encoding.hexbytes import b2h, b2h_rev
from pycoin.serialize import stream_to_bytes
from pycoin.satoshi.checksigops import parse_signature_blob


LOCKTIME_THRESHOLD = 500000000


def dump_header(output, tx):
    tx_bin = stream_to_bytes(tx.stream)
    output.append("Version: %2d  tx hash %s  %d bytes" % (tx.version, tx.id(), len(tx_bin)))
    if tx.has_witness_data():
        output.append("      segwit tx hash %s" % tx.w_id())
    output.append("TxIn count: %d; TxOut count: %d" % (len(tx.txs_in), len(tx.txs_out)))
    if tx.lock_time == 0:
        meaning = "valid anytime"
    elif tx.lock_time < LOCKTIME_THRESHOLD:
        meaning = "valid after block index %d" % tx.lock_time
    else:
        when = datetime.datetime.utcfromtimestamp(tx.lock_time)
        meaning = "valid on or after %s utc" % when.isoformat()
    output.append("Lock time: %d (%s)" % (tx.lock_time, meaning))
    output.append("Input%s:" % ('s' if len(tx.txs_in) != 1 else ''))


def make_trace_script(network, output, do_trace, use_pdb):
    if not (do_trace or use_pdb):
        return None

    def trace_script(opcode, data, pc, vmc):
        output.append("stack: [%s]" % ' '.join(b2h(s) for s in vmc.stack))
        if len(vmc.altstack) > 0:
            output.append("altstack: %s" % vmc.altstack)
        output.append("condition stack: %s" % vmc.conditional_stack)
        output.append("%3d : %02x  %s" % (
            vmc.pc, opcode, network.script.disassemble_for_opcode_data(opcode, data)))
        if use_pdb:
            for line in output:
                print(line)
            output[:] = []
            import pdb
            pdb.set_trace()
    return trace_script


def dump_inputs(output, tx, network, verbose_signature, traceback_f, disassembly_level):
    for idx, tx_in in enumerate(tx.txs_in):
        if tx.is_coinbase():
            output.append("%4d: COINBASE  %12.5f m%s" % (idx, satoshi_to_mbtc(tx.total_in()), network.symbol))
            continue
        suffix = ""
        if tx.missing_unspent(idx):
            tx_out = None
            address = tx_in.address(address_api=network.address)
        else:
            tx_out = tx.unspents[idx]
            sig_result = " sig ok" if tx.is_solution_ok(idx, traceback_f=traceback_f) else " BAD SIG"
            suffix = " %12.5f m%s %s" % (satoshi_to_mbtc(tx_out.coin_value), network.symbol, sig_result)
            address = network.address.for_script(tx_out.puzzle_script())
        t = "%4d: %34s from %s:%-4d%s" % (idx, address, b2h_rev(tx_in.previous_hash),
                                          tx_in.previous_index, suffix)
        output.append(t.rstrip())
        if disassembly_level > 0:
            dump_disassembly(output, tx, idx, network.annotate)

        if verbose_signature:
            dump_signatures(output, tx, tx_in, tx_out, idx, network, traceback_f)


def dump_disassembly(output, tx, tx_in_idx, annotate):
    for (pre_annotations, pc, opcode, instruction, post_annotations) in \
            annotate.annotate_scripts(tx, tx_in_idx):
        for l in pre_annotations:
            output.append("           %s" % l)
        if 1:
            output.append("    %4x: %02x  %s" % (pc, opcode, instruction))
        for l in post_annotations:
            output.append("           %s" % l)


def dump_signatures(output, tx, tx_in, tx_out, idx, network, traceback_f):
    sc = tx.SolutionChecker(tx)
    signatures = [parse_signature_blob(blob) for blob, sig_hash in network.who_signed.extract_signatures(tx, idx)]
    if signatures:
        sig_types_identical = (
            tuple(zip(*signatures))[1].count(signatures[0][1]) == len(signatures))
        i = 1 if len(signatures) > 1 else ''
        for sig_pair, sig_type in signatures:
            output.append("      r{0}: {1:#x}\n      s{0}: {2:#x}".format(i, *sig_pair))
            if not sig_types_identical and tx_out:
                output.append("      z{}: {:#x} {}".format(i, sc._signature_hash(tx_out.script, idx, sig_type),
                              network.annotate.sighash_type_to_string(sig_type)))
            if i:
                i += 1
        if sig_types_identical and tx_out:
            output.append("      z:{} {:#x} {}".format(' ' if i else '', sc._signature_hash(
                tx_out.script, idx, sig_type), network.annotate.sighash_type_to_string(sig_type)))


def dump_footer(network, output, tx, missing_unspents):
    if not missing_unspents:
        output.append("Total input  %12.5f m%s" % (satoshi_to_mbtc(tx.total_in()), network.symbol))
    if 1:
        output.append("Total output %12.5f m%s" % (satoshi_to_mbtc(tx.total_out()), network.symbol))
    if not missing_unspents:
        output.append("Total fees   %12.5f m%s" % (satoshi_to_mbtc(tx.fee()), network.symbol))


def dump_tx(output, tx, network, verbose_signature, disassembly_level, do_trace, use_pdb):
    missing_unspents = tx.missing_unspents()
    traceback_f = make_trace_script(network, output, do_trace, use_pdb)

    dump_header(output, tx)

    dump_inputs(output, tx, network, verbose_signature, traceback_f, disassembly_level)

    output.append("Output%s:" % ('s' if len(tx.txs_out) != 1 else ''))
    for idx, tx_out in enumerate(tx.tx_outs_as_spendable()):
        amount_mbtc = satoshi_to_mbtc(tx_out.coin_value)
        address = network.address.for_script(tx_out.puzzle_script()) or "(unknown)"
        output.append("%4d: %34s receives %12.5f m%s" % (idx, address, amount_mbtc, network.symbol))
        if disassembly_level > 0:
            for (pre_annotations, pc, opcode, instruction, post_annotations) in \
                    network.annotate.annotate_spendable(tx.__class__, tx_out):
                for l in pre_annotations:
                    output.append("           %s" % l)
                if 1:
                    output.append("    %4x: %02x  %s" % (pc, opcode, instruction))
                for l in post_annotations:
                    output.append("           %s" % l)

    dump_footer(network, output, tx, missing_unspents)
